# -*- coding: utf-8 -*-
# Run test_umsgpack.py with your Python interpreter of choice to test the
# correctness of u-msgpack-python!
#
#   $ python2 test_umsgpack.py
#   $ python3 test_umsgpack.py
#   $ pypy test_umsgpack.py
#   $ pypy3 test_umsgpack.py
#

import sys
import struct
import unittest
import datetime
import io
from collections import OrderedDict, namedtuple

import waveforms.server.umsgpack as umsgpack

single_test_vectors = [
    # None
    ["nil", None, b"\xc0"],
    # Booleans
    ["bool false", False, b"\xc2"],
    ["bool true", True, b"\xc3"],
    # + 7-bit uint
    ["7-bit uint", 0x00, b"\x00"],
    ["7-bit uint", 0x10, b"\x10"],
    ["7-bit uint", 0x7f, b"\x7f"],
    # - 5-bit int
    ["5-bit sint", -1, b"\xff"],
    ["5-bit sint", -16, b"\xf0"],
    ["5-bit sint", -32, b"\xe0"],
    # 8-bit uint
    ["8-bit uint", 0x80, b"\xcc\x80"],
    ["8-bit uint", 0xf0, b"\xcc\xf0"],
    ["8-bit uint", 0xff, b"\xcc\xff"],
    # 16-bit uint
    ["16-bit uint", 0x100, b"\xcd\x01\x00"],
    ["16-bit uint", 0x2000, b"\xcd\x20\x00"],
    ["16-bit uint", 0xffff, b"\xcd\xff\xff"],
    # 32-bit uint
    ["32-bit uint", 0x10000, b"\xce\x00\x01\x00\x00"],
    ["32-bit uint", 0x200000, b"\xce\x00\x20\x00\x00"],
    ["32-bit uint", 0xffffffff, b"\xce\xff\xff\xff\xff"],
    # 64-bit uint
    ["64-bit uint", 0x100000000, b"\xcf\x00\x00\x00\x01\x00\x00\x00\x00"],
    ["64-bit uint", 0x200000000000, b"\xcf\x00\x00\x20\x00\x00\x00\x00\x00"],
    ["64-bit uint", 0xffffffffffffffff, b"\xcf\xff\xff\xff\xff\xff\xff\xff\xff"],
    # 8-bit int
    ["8-bit int", -33, b"\xd0\xdf"],
    ["8-bit int", -100, b"\xd0\x9c"],
    ["8-bit int", -128, b"\xd0\x80"],
    # 16-bit int
    ["16-bit int", -129, b"\xd1\xff\x7f"],
    ["16-bit int", -2000, b"\xd1\xf8\x30"],
    ["16-bit int", -32768, b"\xd1\x80\x00"],
    # 32-bit int
    ["32-bit int", -32769, b"\xd2\xff\xff\x7f\xff"],
    ["32-bit int", -1000000000, b"\xd2\xc4\x65\x36\x00"],
    ["32-bit int", -2147483648, b"\xd2\x80\x00\x00\x00"],
    # 64-bit int
    ["64-bit int", -2147483649, b"\xd3\xff\xff\xff\xff\x7f\xff\xff\xff"],
    ["64-bit int", -1000000000000000002, b"\xd3\xf2\x1f\x49\x4c\x58\x9b\xff\xfe"],
    ["64-bit int", -9223372036854775808, b"\xd3\x80\x00\x00\x00\x00\x00\x00\x00"],
    # 64-bit float
    ["64-bit float", 0.0, b"\xcb\x00\x00\x00\x00\x00\x00\x00\x00"],
    ["64-bit float", 2.5, b"\xcb\x40\x04\x00\x00\x00\x00\x00\x00"],
    ["64-bit float", float(10**35), b"\xcb\x47\x33\x42\x61\x72\xc7\x4d\x82"],
    # Fixstr String
    ["fix string", u"", b"\xa0"],
    ["fix string", u"a", b"\xa1\x61"],
    ["fix string", u"abc", b"\xa3\x61\x62\x63"],
    ["fix string", u"a" * 31, b"\xbf" + b"\x61" * 31],
    # 8-bit String
    ["8-bit string", u"b" * 32, b"\xd9\x20" + b"b" * 32],
    ["8-bit string", u"c" * 100, b"\xd9\x64" + b"c" * 100],
    ["8-bit string", u"d" * 255, b"\xd9\xff" + b"d" * 255],
    # 16-bit String
    ["16-bit string", u"b" * 256, b"\xda\x01\x00" + b"b" * 256],
    ["16-bit string", u"c" * 65535, b"\xda\xff\xff" + b"c" * 65535],
    # 32-bit String
    ["32-bit string", u"b" * 65536, b"\xdb\x00\x01\x00\x00" + b"b" * 65536],
    # Wide character String
    ["wide char string", u"Allagbé", b"\xa8Allagb\xc3\xa9"],
    ["wide char string", u"По оживлённым берегам",
        b"\xd9\x28\xd0\x9f\xd0\xbe\x20\xd0\xbe\xd0\xb6\xd0\xb8\xd0\xb2\xd0\xbb\xd1\x91\xd0\xbd\xd0\xbd\xd1\x8b\xd0\xbc\x20\xd0\xb1\xd0\xb5\xd1\x80\xd0\xb5\xd0\xb3\xd0\xb0\xd0\xbc"],
    # 8-bit Binary
    ["8-bit binary", b"\x80" * 1, b"\xc4\x01" + b"\x80" * 1],
    ["8-bit binary", b"\x80" * 32, b"\xc4\x20" + b"\x80" * 32],
    ["8-bit binary", b"\x80" * 255, b"\xc4\xff" + b"\x80" * 255],
    # 16-bit Binary
    ["16-bit binary", b"\x80" * 256, b"\xc5\x01\x00" + b"\x80" * 256],
    # 32-bit Binary
    ["32-bit binary", b"\x80" * 65536, b"\xc6\x00\x01\x00\x00" + b"\x80" * 65536],
    # Fixext 1
    ["fixext 1", umsgpack.Ext(0x05, b"\x80" * 1), b"\xd4\x05" + b"\x80" * 1],
    # Fixext 2
    ["fixext 2", umsgpack.Ext(0x05, b"\x80" * 2), b"\xd5\x05" + b"\x80" * 2],
    # Fixext 4
    ["fixext 4", umsgpack.Ext(0x05, b"\x80" * 4), b"\xd6\x05" + b"\x80" * 4],
    # Fixext 8
    ["fixext 8", umsgpack.Ext(0x05, b"\x80" * 8), b"\xd7\x05" + b"\x80" * 8],
    # Fixext 16
    ["fixext 16", umsgpack.Ext(0x05, b"\x80" * 16),
        b"\xd8\x05" + b"\x80" * 16],
    # 8-bit Ext
    ["8-bit ext", umsgpack.Ext(0x05, b"\x80" * 255),
        b"\xc7\xff\x05" + b"\x80" * 255],
    # 16-bit Ext
    ["16-bit ext", umsgpack.Ext(0x05, b"\x80" * 256),
        b"\xc8\x01\x00\x05" + b"\x80" * 256],
    # 32-bit Ext
    ["32-bit ext", umsgpack.Ext(0x05, b"\x80" * 65536),
        b"\xc9\x00\x01\x00\x00\x05" + b"\x80" * 65536],
    # Empty Array
    ["empty array", [], b"\x90"],
    # Empty Map
    ["empty map", {}, b"\x80"],
    # 32-bit Timestamp
    ["32-bit timestamp", datetime.datetime(1970, 1, 1, 0, 0, 0, 0, umsgpack._utc_tzinfo),
        b"\xd6\xff\x00\x00\x00\x00"],
    ["32-bit timestamp", datetime.datetime(2000, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo),
        b"\xd6\xff\x38\x6d\xd1\x4e"],
    # 64-bit Timestamp
    ["64-bit timestamp", datetime.datetime(2000, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo),
        b"\xd7\xff\x00\x4b\x51\x40\x38\x6d\xd1\x4e"],
    ["64-bit timestamp", datetime.datetime(2200, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo),
        b"\xd7\xff\x00\x00\x00\x01\xb0\x9e\xa6\xce"],
    ["64-bit timestamp", datetime.datetime(2200, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo),
        b"\xd7\xff\x00\x4b\x51\x41\xb0\x9e\xa6\xce"],
    # 96-bit Timestamp
    ["96-bit timestamp", datetime.datetime(1900, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo),
        b"\xc7\x0c\xff\x00\x00\x00\x00\xff\xff\xff\xff\x7c\x56\x0f\x4e"],
    ["96-bit timestamp", datetime.datetime(1900, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo),
        b"\xc7\x0c\xff\x00\x12\xd4\x50\xff\xff\xff\xff\x7c\x56\x0f\x4e"],
    ["96-bit timestamp", datetime.datetime(3000, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo),
        b"\xc7\x0c\xff\x00\x00\x00\x00\x00\x00\x00\x07\x91\x5f\x59\xce"],
    ["96-bit timestamp", datetime.datetime(3000, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo),
        b"\xc7\x0c\xff\x00\x12\xd4\x50\x00\x00\x00\x07\x91\x5f\x59\xce"],
]

composite_test_vectors = [
    # Fix Array
    ["fix array", [5, u"abc", True],
        b"\x93\x05\xa3\x61\x62\x63\xc3"],
    # 16-bit Array
    ["16-bit array", [0x05] * 16,
        b"\xdc\x00\x10" + b"\x05" * 16],
    ["16-bit array", [0x05] * 65535,
        b"\xdc\xff\xff" + b"\x05" * 65535],
    # 32-bit Array
    ["32-bit array", [0x05] * 65536,
        b"\xdd\x00\x01\x00\x00" + b"\x05" * 65536],
    # Fix Map
    ["fix map", OrderedDict([(1, True), (2, u"abc"), (3, b"\x80")]),
        b"\x83\x01\xc3\x02\xa3\x61\x62\x63\x03\xc4\x01\x80"],
    ["fix map", {u"abc": 5},
        b"\x81\xa3\x61\x62\x63\x05"],
    ["fix map", {b"\x80": 0xffff},
        b"\x81\xc4\x01\x80\xcd\xff\xff"],
    ["fix map", {True: None},
        b"\x81\xc3\xc0"],
    # 16-bit Map
    ["16-bit map", OrderedDict([(k, 0x05) for k in range(16)]),
        b"\xde\x00\x10" + b"".join([struct.pack("B", i) + b"\x05" for i in range(16)])],
    ["16-bit map", OrderedDict([(k, 0x05) for k in range(6000)]),
        b"\xde\x17\x70" + b"".join([struct.pack("B", i) + b"\x05" for i in range(128)]) +
        b"".join([b"\xcc" + struct.pack("B", i) + b"\x05" for i in range(128, 256)]) +
        b"".join([b"\xcd" + struct.pack(">H", i) + b"\x05" for i in range(256, 6000)])],
    # Complex Array
    ["complex array", [True, 0x01, umsgpack.Ext(0x03, b"foo"), 0xff,
                       OrderedDict([(1, False), (2, u"abc")]), b"\x80",
                       [1, 2, 3], u"abc"],
        b"\x98\xc3\x01\xc7\x03\x03\x66\x6f\x6f\xcc\xff\x82\x01\xc2\x02\xa3\x61\x62\x63\xc4\x01\x80\x93\x01\x02\x03\xa3\x61\x62\x63"],
    # Complex Map
    ["complex map", OrderedDict([(1, [OrderedDict([(1, 2), (3, 4)]), {}]),
                                 (2, 1), (3, [False, u"def"]),
                                 (4, OrderedDict([(0x100000000, u"a"),
                                                  (0xffffffff, u"b")]))]),
        b"\x84\x01\x92\x82\x01\x02\x03\x04\x80\x02\x01\x03\x92\xc2\xa3\x64\x65\x66\x04\x82\xcf\x00\x00\x00\x01\x00\x00\x00\x00\xa1\x61\xce\xff\xff\xff\xff\xa1\x62"],
    # Map with Tuple Keys
    ["map with tuple keys", OrderedDict([((u"foo", False, 3), True),
                                         ((3e6, -5), u"def")]),
        b"\x82\x93\xa3\x66\x6f\x6f\xc2\x03\xc3\x92\xcb\x41\x46\xe3\x60\x00\x00\x00\x00\xfb\xa3\x64\x65\x66"],
    # Map with Complex Tuple Keys
    ["map with complex tuple keys", {(u"foo", (1, 2, 3), 3): -5},
        b"\x81\x93\xa3\x66\x6f\x6f\x93\x01\x02\x03\x03\xfb"]
]

pack_exception_test_vectors = [
    # Unsupported type exception
    ["unsupported type", set([1, 2, 3]), umsgpack.UnsupportedTypeException],
    ["unsupported type", -2**(64 - 1) - 1, umsgpack.UnsupportedTypeException],
    ["unsupported type", 2**64, umsgpack.UnsupportedTypeException],
]

unpack_exception_test_vectors = [
    # Type errors
    ["type error unpack unicode string", u"\x01", TypeError],
    ["type error unpack boolean", True, TypeError],
    # Insufficient data to unpack object
    ["insufficient data 8-bit uint", b"\xcc",
        umsgpack.InsufficientDataException],
    ["insufficient data 16-bit uint", b"\xcd\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit uint", b"\xce\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 64-bit uint", b"\xcf\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 8-bit int", b"\xd0",
        umsgpack.InsufficientDataException],
    ["insufficient data 16-bit int", b"\xd1\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit int", b"\xd2\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 64-bit int", b"\xd3\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit float", b"\xca\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data 64-bit float", b"\xcb\xff",
        umsgpack.InsufficientDataException],
    ["insufficient data fixstr", b"\xa1",
        umsgpack.InsufficientDataException],
    ["insufficient data 8-bit string", b"\xd9",
        umsgpack.InsufficientDataException],
    ["insufficient data 8-bit string", b"\xd9\x01",
        umsgpack.InsufficientDataException],
    ["insufficient data 16-bit string", b"\xda\x01\x00",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit string", b"\xdb\x00\x01\x00\x00",
        umsgpack.InsufficientDataException],
    ["insufficient data 8-bit binary", b"\xc4",
        umsgpack.InsufficientDataException],
    ["insufficient data 8-bit binary", b"\xc4\x01",
        umsgpack.InsufficientDataException],
    ["insufficient data 16-bit binary", b"\xc5\x01\x00",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit binary", b"\xc6\x00\x01\x00\x00",
        umsgpack.InsufficientDataException],
    ["insufficient data fixarray", b"\x91",
        umsgpack.InsufficientDataException],
    ["insufficient data fixarray", b"\x92\xc2",
        umsgpack.InsufficientDataException],
    ["insufficient data 16-bit array", b"\xdc\x00\xf0\xc2\xc3",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit array", b"\xdd\x00\x01\x00\x00\xc2\xc3",
        umsgpack.InsufficientDataException],
    ["insufficient data fixmap", b"\x81",
        umsgpack.InsufficientDataException],
    ["insufficient data fixmap", b"\x82\xc2\xc3",
        umsgpack.InsufficientDataException],
    ["insufficient data 16-bit map", b"\xde\x00\xf0\xc2\xc3",
        umsgpack.InsufficientDataException],
    ["insufficient data 32-bit map", b"\xdf\x00\x01\x00\x00\xc2\xc3",
        umsgpack.InsufficientDataException],
    ["insufficient data fixext 1", b"\xd4",
        umsgpack.InsufficientDataException],
    ["insufficient data fixext 1", b"\xd4\x05",
        umsgpack.InsufficientDataException],
    ["insufficient data fixext 2", b"\xd5\x05\x01",
        umsgpack.InsufficientDataException],
    ["insufficient data fixext 4", b"\xd6\x05\x01\x02\x03",
        umsgpack.InsufficientDataException],
    ["insufficient data fixext 8", b"\xd7\x05\x01\x02\x03",
        umsgpack.InsufficientDataException],
    ["insufficient data fixext 16", b"\xd8\x05\x01\x02\x03",
        umsgpack.InsufficientDataException],
    ["insufficient data ext 8-bit", b"\xc7\x05\x05\x01\x02\x03",
        umsgpack.InsufficientDataException],
    ["insufficient data ext 16-bit", b"\xc8\x01\x00\x05\x01\x02\x03",
        umsgpack.InsufficientDataException],
    ["insufficient data ext 32-bit", b"\xc9\x00\x01\x00\x00\x05\x01\x02\x03",
        umsgpack.InsufficientDataException],
    # Unhashable key { 1 : True, { 1 : 1 } : False }
    ["unhashable key", b"\x82\x01\xc3\x81\x01\x01\xc2",
        umsgpack.UnhashableKeyException],
    # Unhashable key { [ 1, 2, {} ] : True }
    ["unhashable key", b"\x81\x93\x01\x02\x80\xc3",
        umsgpack.UnhashableKeyException],
    # Key duplicate { 1 : True, 1 : False }
    ["duplicate key", b"\x82\x01\xc3\x01\xc2",
        umsgpack.DuplicateKeyException],
    # Reserved code (0xc1)
    ["reserved code", b"\xc1",
        umsgpack.ReservedCodeException],
    # Unsupported timestamp (unsupported data length)
    ["unsupported timestamp", b"\xc7\x02\xff\xaa\xbb",
        umsgpack.UnsupportedTimestampException],
    # Invalid string (non utf-8)
    ["invalid string", b"\xa1\x80",
        umsgpack.InvalidStringException],
]

compatibility_test_vectors = [
    # Fix Raw
    ["fix raw", b"", b"\xa0"],
    ["fix raw", u"", b"\xa0"],
    ["fix raw", b"a", b"\xa1\x61"],
    ["fix raw", u"a", b"\xa1\x61"],
    ["fix raw", b"abc", b"\xa3\x61\x62\x63"],
    ["fix raw", u"abc", b"\xa3\x61\x62\x63"],
    ["fix raw", b"a" * 31, b"\xbf" + b"\x61" * 31],
    ["fix raw", u"a" * 31, b"\xbf" + b"\x61" * 31],
    # 16-bit Raw
    ["16-bit raw", u"b" * 32, b"\xda\x00\x20" + b"b" * 32],
    ["16-bit raw", b"b" * 32, b"\xda\x00\x20" + b"b" * 32],
    ["16-bit raw", u"b" * 256, b"\xda\x01\x00" + b"b" * 256],
    ["16-bit raw", b"b" * 256, b"\xda\x01\x00" + b"b" * 256],
    ["16-bit raw", u"c" * 65535, b"\xda\xff\xff" + b"c" * 65535],
    ["16-bit raw", b"c" * 65535, b"\xda\xff\xff" + b"c" * 65535],
    # 32-bit Raw
    ["32-bit raw", u"b" * 65536, b"\xdb\x00\x01\x00\x00" + b"b" * 65536],
    ["32-bit raw", b"b" * 65536, b"\xdb\x00\x01\x00\x00" + b"b" * 65536],
]

float_precision_test_vectors = [
    ["float precision single", 2.5, b"\xca\x40\x20\x00\x00"],
    ["float precision double", 2.5, b"\xcb\x40\x04\x00\x00\x00\x00\x00\x00"],
]

tuple_test_vectors = [
    ["nested array", [0x01, [b"\x80", [[u"a", u"b", u"c"], True]]],
        b"\x92\x01\x92\xc4\x01\x80\x92\x93\xa1a\xa1b\xa1c\xc3",
        (0x01, (b"\x80", ((u"a", u"b", u"c"), True)))],
]

naive_timestamp_test_vectors = [
    ["32-bit timestamp (naive)", datetime.datetime(2000, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo),
        b"\xd6\xff\x38\x6d\xd1\x4e",
        datetime.datetime(2000, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo)],
    ["64-bit timestamp (naive)", datetime.datetime(2200, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo),
        b"\xd7\xff\x00\x4b\x51\x41\xb0\x9e\xa6\xce",
        datetime.datetime(2200, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo)],
    ["96-bit timestamp (naive)", datetime.datetime(3000, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo),
        b"\xc7\x0c\xff\x00\x12\xd4\x50\x00\x00\x00\x07\x91\x5f\x59\xce",
        datetime.datetime(3000, 1, 1, 10, 5, 2, 1234, umsgpack._utc_tzinfo)],
]

CustomType = namedtuple('CustomType', ['x', 'y', 'z'])

ext_handlers = {
    complex: lambda obj: umsgpack.Ext(0x20, struct.pack("<ff", obj.real, obj.imag)),
    CustomType: lambda obj: umsgpack.Ext(0x30, umsgpack.packb(list(obj))),
    0x20: lambda ext: complex(*struct.unpack("<ff", ext.data)),
    0x30: lambda ext: CustomType(*umsgpack.unpackb(ext.data)),
}

ext_handlers_test_vectors = [
    ["complex", complex(1, 2), b"\xd7\x20\x00\x00\x80\x3f\x00\x00\x00\x40"],
    ["custom type", CustomType(b"abc", 123, True),
     b"\xd7\x30\x93\xc4\x03\x61\x62\x63\x7b\xc3"],
]

override_ext_handlers = {
    datetime.datetime:
        lambda obj: umsgpack.Ext(0x40, obj.strftime("%Y%m%dT%H:%M:%S.%f").encode()),
    -0x01:
        lambda ext: ext,
}

override_ext_handlers_test_vectors = [
    ["pack override",
        datetime.datetime(2000, 1, 1, 10, 5, 2, 0, umsgpack._utc_tzinfo),
        b'\xc7\x18@20000101T10:05:02.000000'],
    ["unpack override",
        umsgpack.Ext(-0x01, b"\x00\xbb\xcc\xdd\x01\x02\x03\x04\x05\x06\x07\x08"),
        b'\xc7\x0c\xff\x00\xbb\xcc\xdd\x01\x02\x03\x04\x05\x06\x07\x08'],
]

# These are the only global variables that should be exported by umsgpack
exported_vars_test_vector = [
    "Ext",
    "InvalidString",
    "PackException",
    "UnpackException",
    "UnsupportedTypeException",
    "InsufficientDataException",
    "InvalidStringException",
    "UnsupportedTimestampException",
    "ReservedCodeException",
    "UnhashableKeyException",
    "DuplicateKeyException",
    "KeyNotPrimitiveException",
    "KeyDuplicateException",
    "ext_serializable",
    "pack",
    "packb",
    "unpack",
    "unpackb",
    "dump",
    "dumps",
    "load",
    "loads",
    "version",
    "compatibility",
]

##########################################################################


class TestUmsgpack(unittest.TestCase):

    def test_pack_single(self):
        for (name, obj, data) in single_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            self.assertEqual(umsgpack.packb(obj), data)

    def test_pack_composite(self):
        for (name, obj, data) in composite_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            self.assertEqual(umsgpack.packb(obj), data)

    def test_pack_exceptions(self):
        for (name, obj, exception) in pack_exception_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            with self.assertRaises(exception):
                umsgpack.packb(obj)

    def test_unpack_single(self):
        for (name, obj, data) in single_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            unpacked = umsgpack.unpackb(data)

            # In Python2, we have both int and long integer types, but which
            # one we end up with depends on the architecture (32-bit/64-bit)
            if sys.version_info[0] == 2:
                # Allow both {int,long} -> unpackb -> {int,long}
                if isinstance(obj, int) or isinstance(obj, long):
                    self.assertTrue(isinstance(unpacked, int) or
                                    isinstance(unpacked, long))
                else:
                    self.assertTrue(isinstance(unpacked, type(obj)))
            # In Python3, we only have the int integer type
            else:
                self.assertTrue(isinstance(unpacked, type(obj)))

            self.assertEqual(unpacked, obj)

    def test_unpack_composite(self):
        for (name, obj, data) in composite_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            self.assertEqual(umsgpack.unpackb(data), obj)

    def test_unpack_exceptions(self):
        for (name, data, exception) in unpack_exception_test_vectors:
            print("\tTesting {:s}".format(name))

            with self.assertRaises(exception):
                umsgpack.unpackb(data)

    def test_pack_compatibility(self):
        umsgpack.compatibility = True

        for (name, obj, data) in compatibility_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            self.assertEqual(umsgpack.packb(obj), data)

        umsgpack.compatibility = False

    def test_unpack_compatibility(self):
        umsgpack.compatibility = True

        for (name, obj, data) in compatibility_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            unpacked = umsgpack.unpackb(data)

            # Encoded raw should always unpack to bytes in compatibility mode,
            # so convert any string obj to bytes before comparison
            if sys.version_info[0] == 3 and isinstance(obj, str):
                _obj = obj.encode('utf-8')
            else:
                _obj = obj

            self.assertTrue(isinstance(unpacked, type(_obj)))
            self.assertEqual(unpacked, _obj)

        umsgpack.compatibility = False

    def test_unpack_invalid_string(self):
        # Use last unpack exception test vector (an invalid string)
        (_, data, _) = unpack_exception_test_vectors[-1]

        obj = umsgpack.unpackb(data, allow_invalid_utf8=True)
        self.assertTrue(isinstance(obj, umsgpack.InvalidString))
        self.assertEqual(obj, b"\x80")

    def test_unpack_ordered_dict(self):
        # Use last composite test vector (a map)
        (_, obj, data) = composite_test_vectors[-1]

        # Unpack with default options (unordered dict)
        unpacked = umsgpack.unpackb(data)
        self.assertTrue(isinstance(unpacked, dict))

        # Unpack with unordered dict
        unpacked = umsgpack.unpackb(data, use_ordered_dict=False)
        self.assertTrue(isinstance(unpacked, dict))

        # Unpack with ordered dict
        unpacked = umsgpack.unpackb(data, use_ordered_dict=True)
        self.assertTrue(isinstance(unpacked, OrderedDict))
        self.assertEqual(unpacked, obj)

    def test_unpack_tuple(self):
        # Use tuple test vector
        (_, obj, data, obj_tuple) = tuple_test_vectors[0]

        # Unpack with default options (list)
        self.assertEqual(umsgpack.unpackb(data), obj)

        # Unpack with use_tuple=False (list)
        self.assertEqual(umsgpack.unpackb(data, use_tuple=False), obj)

        # Unpack with use_tuple=True (tuple)
        self.assertEqual(umsgpack.unpackb(data, use_tuple=True), obj_tuple)

    def test_ext_exceptions(self):
        # Test invalid Ext type type
        with self.assertRaises(TypeError):
            _ = umsgpack.Ext(5.0, b"")

        # Test invalid data type
        with self.assertRaises(TypeError):
            _ = umsgpack.Ext(0, u"unicode string")

        # Test out of range Ext type value
        with self.assertRaises(ValueError):
            _ = umsgpack.Ext(-129, b"data")
        with self.assertRaises(ValueError):
            _ = umsgpack.Ext(128, b"data")

    def test_pack_ext_handler(self):
        for (name, obj, data) in ext_handlers_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            packed = umsgpack.packb(obj, ext_handlers=ext_handlers)
            self.assertEqual(packed, data)

    def test_unpack_ext_handler(self):
        for (name, obj, data) in ext_handlers_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            unpacked = umsgpack.unpackb(data, ext_handlers=ext_handlers)
            self.assertEqual(unpacked, obj)

    def test_pack_force_float_precision(self):
        for ((name, obj, data), precision) in zip(float_precision_test_vectors, ["single", "double"]):
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            packed = umsgpack.packb(obj, force_float_precision=precision)
            self.assertEqual(packed, data)

    def test_pack_naive_timestamp(self):
        for (name, obj, data, _) in naive_timestamp_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            packed = umsgpack.packb(obj)
            self.assertEqual(packed, data)

    def test_unpack_naive_timestamp(self):
        for (name, _, data, obj) in naive_timestamp_test_vectors:
            obj_repr = repr(obj)
            print("\tTesting {:s}: object {:s}".format(
                  name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

            unpacked = umsgpack.unpackb(data)
            self.assertEqual(unpacked, obj)

    def test_pack_ext_override(self):
        # Test overridden packing of datetime.datetime
        (name, obj, data) = override_ext_handlers_test_vectors[0]
        obj_repr = repr(obj)
        print("\tTesting {:s}: object {:s}".format(
              name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

        packed = umsgpack.packb(obj, ext_handlers=override_ext_handlers)
        self.assertEqual(packed, data)

    def test_unpack_ext_override(self):
        # Test overridden unpacking of Ext type -1
        (name, obj, data) = override_ext_handlers_test_vectors[1]
        obj_repr = repr(obj)
        print("\tTesting {:s}: object {:s}".format(
              name, obj_repr if len(obj_repr) < 24 else obj_repr[0:24] + "..."))

        unpacked = umsgpack.unpackb(data, ext_handlers=override_ext_handlers)
        self.assertEqual(unpacked, obj)

    def test_ext_handlers_subclass(self):
        class Rectangle:
            def __init__(self, length, width):
                self.length = length
                self.width = width

            def __eq__(self, other):
                return self.length == other.length and self.width == other.width

        class Square(Rectangle):
            def __init__(self, width):
                Rectangle.__init__(self, width, width)

        # Test pack (packs base class)
        packed = umsgpack.packb(Square(5), ext_handlers={
            Rectangle: lambda obj: umsgpack.Ext(0x10, umsgpack.packb([obj.length, obj.width])),
        })
        self.assertEqual(packed, b"\xc7\x03\x10\x92\x05\x05")

        # Test unpack (unpacks base class)
        unpacked = umsgpack.unpackb(packed, ext_handlers={
            0x10: lambda ext: Rectangle(*umsgpack.unpackb(ext.data)),
        })
        self.assertEqual(unpacked, Rectangle(5, 5))

    def test_ext_serializable(self):
        # Register test class
        @umsgpack.ext_serializable(0x20)
        class CustomComplex:
            def __init__(self, real, imag):
                self.real = real
                self.imag = imag

            def __eq__(self, other):
                return self.real == other.real and self.imag == other.imag

            def packb(self):
                return struct.pack("<II", self.real, self.imag)

            @classmethod
            def unpackb(cls, data):
                return cls(*struct.unpack("<II", data))

        obj, data = CustomComplex(123, 456), b"\xd7\x20\x7b\x00\x00\x00\xc8\x01\x00\x00"

        # Test pack
        packed = umsgpack.packb(obj)
        self.assertEqual(packed, data)

        # Test unpack
        unpacked = umsgpack.unpackb(packed)
        self.assertTrue(isinstance(unpacked, CustomComplex))
        self.assertEqual(unpacked, obj)

        _, obj, data = ext_handlers_test_vectors[0]

        # Test pack priority of ext_handlers over ext_serializable()
        packed = umsgpack.packb(obj, ext_handlers=ext_handlers)
        self.assertEqual(packed, data)

        # Test unpack priority of ext_handlers over ext_serializable()
        unpacked = umsgpack.unpackb(data, ext_handlers=ext_handlers)
        self.assertTrue(isinstance(unpacked, complex))
        self.assertEqual(unpacked, obj)

        # Test registration collision
        with self.assertRaises(ValueError):
            @umsgpack.ext_serializable(0x20)
            class DummyClass:
                pass

        # Test out of range Ext type value
        with self.assertRaises(ValueError):
            @umsgpack.ext_serializable(-129)
            class DummyClass2:
                pass
        with self.assertRaises(ValueError):
            @umsgpack.ext_serializable(128)
            class DummyClass3:
                pass

        # Register class with missing packb() and unpackb()
        @umsgpack.ext_serializable(0x21)
        class IncompleteClass:
            pass

        # Test unimplemented packb()
        with self.assertRaises(NotImplementedError):
            umsgpack.packb(IncompleteClass())

        # Test unimplemented unpackb()
        with self.assertRaises(NotImplementedError):
            umsgpack.unpackb(b"\xd4\x21\x00")

        # Unregister Ext serializable classes to prevent interference with
        # subsequent tests
        umsgpack._ext_classes_to_code = {}
        umsgpack._ext_code_to_classes = {}

    def test_ext_serializable_subclass(self):
        @umsgpack.ext_serializable(0x10)
        class Rectangle:
            def __init__(self, length, width):
                self.length = length
                self.width = width

            def __eq__(self, other):
                return self.length == other.length and self.width == other.width

            def packb(self):
                return umsgpack.packb([self.length, self.width])

            @classmethod
            def unpackb(cls, data):
                return cls(*umsgpack.unpackb(data))

        class Square(Rectangle):
            def __init__(self, width):
                Rectangle.__init__(self, width, width)

        # Test pack (packs base class)
        packed = umsgpack.packb(Square(5))
        self.assertEqual(packed, b"\xc7\x03\x10\x92\x05\x05")

        # Test unpack (unpacks base class)
        unpacked = umsgpack.unpackb(packed)
        self.assertEqual(unpacked, Rectangle(5, 5))

        # Unregister Ext serializable classes to prevent interference with
        # subsequent tests
        umsgpack._ext_classes_to_code = {}
        umsgpack._ext_code_to_classes = {}

    def test_streaming_writer(self):
        # Try first composite test vector
        (_, obj, data) = composite_test_vectors[0]
        writer = io.BytesIO()
        umsgpack.pack(obj, writer)
        self.assertTrue(writer.getvalue(), data)

    def test_streaming_reader(self):
        # Try first composite test vector
        (_, obj, data) = composite_test_vectors[0]
        reader = io.BytesIO(data)
        self.assertEqual(umsgpack.unpack(reader), obj)

    def test_namespacing(self):
        # Get a list of global variables from umsgpack module
        exported_vars = list([x for x in dir(umsgpack) if not x.startswith("_")])
        # Ignore imports
        exported_vars = list([x for x in exported_vars if x != "struct" and x != "collections" and x != "datetime" and x !=
                                    "sys" and x != "io" and x != "xrange" and x != "Hashable"])

        # self.assertTrue(len(exported_vars) == len(exported_vars_test_vector))
        for var in exported_vars_test_vector:
            self.assertTrue(var in exported_vars)

    def test_load_short_read(self):
        # When reading from files, the network, etc. there's no guarantee that
        # read(n) returns n bytes. Simulate this with a file-like object that
        # returns 1 byte at a time.

        class SlowFile(object):
            def __init__(self, data):
                self._data = data

            def read(self, n=None):
                if n is None or len(self._data) == 0:
                    data, self._data = self._data, b''
                    return data

                chunk = self._data[0:1]
                self._data = self._data[1:]
                return chunk

        obj = {'hello': 'world'}
        f = SlowFile(umsgpack.dumps(obj))
        unpacked = umsgpack.load(f)

        self.assertEqual(unpacked, obj)


if __name__ == '__main__':
    unittest.main()


from .bayes import bayesian_correction
from .fit import (fit_circle, fit_cross_point, fit_pole, goodness_of_fit,
                  lin_fit)
from .prime import Primes
from .signal.demodulate import getFTMatrix



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base node object for the OPENQASM syntax tree."""


class Node:
    """Base node object for the OPENQASM syntax tree."""

    def __init__(self, type, children=None, root=None):
        """Construct a new node object."""
        # pylint: disable=redefined-builtin
        self.type = type
        if children:
            self.children = children
        else:
            self.children = []
        self.root = root
        # True if this node is an expression node, False otherwise
        self.expression = False

    def is_expression(self):
        """Return True if this is an expression node."""
        return self.expression

    def add_child(self, node):
        """Add a child node."""
        self.children.append(node)

    def to_string(self, indent):
        """Print with indent."""
        ind = indent * ' '
        if self.root:
            print(ind, self.type, '---', self.root)
        else:
            print(ind, self.type)
        indent = indent + 3
        ind = indent * ' '
        for children in self.children:
            if children is None:
                print("OOPS! type of parent is", type(self))
                print(self.children)
            if isinstance(children, str):
                print(ind, children)
            elif isinstance(children, int):
                print(ind, str(children))
            elif isinstance(children, float):
                print(ind, str(children))
            else:
                children.to_string(indent)


import warnings
from collections import defaultdict

import numpy as np
from scipy.optimize import minimize
from scipy.special import erf, erfinv

from .geo import EPS, point_in_polygon


def count_state(state):
    ret = defaultdict(lambda: 0)
    for s in state:
        ret[tuple(s)] += 1
    return dict(ret)


def count_to_diag(count, shape=None):
    state = list(count.keys())
    if shape is None:
        shape = (2, ) * len(state[0])
    n = np.asarray(list(count.values()))
    p = n / np.sum(n)
    state = np.ravel_multi_index(np.asarray(state).T, shape)
    ret = np.zeros(shape).reshape(-1)
    ret[state] = p
    return ret


def _atleast_type(a, dtype):
    if dtype == np.double and a.dtype.type == np.int8:
        return a.astype(np.double)
    elif dtype == complex and a.dtype.type in [np.int8, np.double]:
        return a.astype(complex)
    else:
        return a


__classify_methods = {}


def install_classify_method(method: str, func: callable):
    __classify_methods[method] = func


def uninstall_classify_method(method: str):
    if method in __classify_methods:
        del __classify_methods[method]


def classify(data, method, params):
    if method in __classify_methods:
        return __classify_methods[method](data, params)
    else:
        raise ValueError("method not found")


def default_classify(data, params):
    """
    默认的分类方法
    """
    thr = params.get('threshold', 0)
    phi = params.get('phi', 0)
    return (data * np.exp(-1j * phi)).real > thr


def classify_svm(data, params):
    """
    分类方法：SVM
    """
    raise NotImplementedError
    from sklearn import svm

    clf = svm.SVC(kernel='rbf',
                  gamma=params.get('gamma', 1),
                  C=params.get('C', 1))
    clf.fit(data, data)
    return clf.predict(data)


def classify_kmeans(data, params):
    """
    分类方法：KMeans
    """
    from sklearn.cluster import KMeans

    centers = params.get('centers', None)
    if isinstance(centers, list):
        centers = np.asarray(centers)

    k = params.get('k', None)
    if k is None and centers is not None:
        k = np.asarray(centers).shape[0]
    cur_shape = data.shape

    flatten_init = np.array([np.real(centers), np.imag(centers)]).T

    flatten_data = data.flatten()
    ret_ans = KMeans(n_clusters=k, init=flatten_init).fit_predict(
        np.array([np.real(flatten_data),
                  np.imag(flatten_data)]).T)
    return 2**ret_ans.reshape(cur_shape)


def classify_nearest(data, params):
    """
    分类方法：最近邻
    """
    centers = params.get('centers', None)
    if centers is None:
        raise ValueError("centers not found")
    return 2**np.argmin([np.abs(data - c) for c in centers], axis=0)


def classify_range(data, params):
    """
    分类方法：范围
    """
    centers = params.get('centers', None)
    radians = params.get('radians', None)
    if centers is None:
        raise ValueError("centers not found")
    if radians is None:
        return 2**np.argmin([np.abs(data - c) for c in centers], axis=0)

    ret = np.full_like(data, 0, dtype=int)
    for i, (c, r) in enumerate(zip(centers, radians)):
        ret[np.abs(data - c) <= r] += 2**i
    return ret


def classify_polygon(data, params):
    """
    分类方法: 多边形内
    """
    polygons = params.get('polygons', None)
    eps = params.get('eps', EPS)
    if polygons is None:
        raise ValueError("polygons not found")

    ret = np.full_like(data, 0, dtype=int)
    for i, polygon in enumerate(polygons):
        ret[point_in_polygon(data, polygon, eps)] += 2**i
    return ret


install_classify_method("state", default_classify)
install_classify_method("nearest", classify_nearest)
install_classify_method("range", classify_range)
install_classify_method("kmeans", classify_kmeans)
install_classify_method("polygon", classify_polygon)


def classify_data(data, measure_gates, avg=False):
    assert data.shape[-1] == len(
        measure_gates), 'number of qubits must equal to the size of last axis'

    ret = np.zeros_like(data, dtype=np.int8)

    for i, g in enumerate(measure_gates):
        signal = g['params'].get('signal', 'state')

        if signal in __classify_methods:
            ret[..., i] = classify(data[..., i], signal, g['params'])
        elif signal == 'amp':
            phi = g['params'].get('phi', 0)
            ret = _atleast_type(ret, np.double)
            ret[..., i] = (data[..., i] * np.exp(-1j * phi)).real
        if signal == 'raw':
            ret = _atleast_type(ret, complex)
            ret[..., i] = data[..., i]
        elif signal == 'real':
            ret = _atleast_type(ret, np.double)
            ret[..., i] = data[..., i].real
        elif signal == 'imag':
            ret = _atleast_type(ret, np.double)
            ret[..., i] = data[..., i].imag
        elif signal == 'abs':
            ret = _atleast_type(ret, np.double)
            ret[..., i] = np.abs(data[..., i])
        else:
            pass
    if avg:
        ret = ret.mean(axis=-2)
    return ret


def cdf(t, data):
    data.sort()
    x = data
    y = np.linspace(0, 1, len(data))
    if t is None:
        return x, y
    else:
        return np.interp(t, x, y, left=0, right=1)


def gaussian_pdf(x, mu, sigma):
    return 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(-0.5 *
                                                     ((x - mu) / sigma)**2)


def gaussian_pdf_2d(z, mu, cov):
    z = z - mu
    v = np.moveaxis(np.array([z.real, z.imag]), 0, -1).reshape(*z.shape, 2, 1)
    vT = np.moveaxis(v, -1, -2)
    m = np.linalg.inv(cov)
    return 1 / (2 * np.pi * np.sqrt(np.linalg.det(cov))) * np.exp(
        -0.5 * vT @ m @ v)[..., 0, 0]


def gaussian_cdf(x, mu, sigma):
    return 0.5 * (1 + erf((x - mu) / (sigma * np.sqrt(2))))


def gaussian_cdf_inv(y, mu, sigma):
    return np.sqrt(2) * sigma * erfinv(2 * y - 1) + mu


def mult_gaussian_pdf(x, mu, sigma, amp):
    amp /= np.sum(amp)
    ret = np.zeros_like(x)
    for i in range(len(mu)):
        ret += amp[i] * gaussian_pdf(x, mu[i], sigma[i])
    return ret


def mult_gaussian_cdf(x, mu, sigma, amp):
    amp /= np.sum(amp)
    ret = np.zeros_like(x)
    for i in range(len(mu)):
        ret += amp[i] * gaussian_cdf(x, mu[i], sigma[i])
    return ret


def readout_distribution(s, p, c0, c1, cov0, cov1):
    return gaussian_pdf_2d(s, c0, cov0) * p + gaussian_pdf_2d(s, c1,
                                                              cov1) * (1 - p)


def fit_readout_distribution(s0, s1):
    center = 0.5 * (s0.mean() + s1.mean())
    s0, s1 = s0 - center, s1 - center

    scale = np.max([np.abs(s0.mean()), np.abs(s1.mean()), s0.std(), s1.std()])

    s0, s1 = s0 / scale, s1 / scale

    def a_b_phi_2_cov(a, b, phi):
        m = np.array([[np.cos(phi) * a, -np.sin(phi) * b],
                      [np.sin(phi) * a, np.cos(phi) * b]])

        return m @ m.T

    def cov_2_a_b_phi(cov):
        x, y, z = cov[0, 0], cov[1, 1], cov[0, 1]
        a = (x - y) / z
        b = np.sqrt(4 + a**2)
        c = np.sqrt(8 + 2 * a**2 + (2 * a**3 + 8 * a) / b)
        t = -a / 2 - b / 2 + c / 2
        d = 1 - 6 * t**2 + t**4
        phi = np.arctan2(2 * t, 1 - t**2)
        a = np.sqrt(-(-x + 2 * t**2 * x - t**4 * x + 4 * t**2 * y) / d)
        b = np.sqrt(-(4 * t**2 * x - y + 2 * t**2 * y - t**4 * y) / d)
        return a, b, phi

    def loss(params, s0, s1):
        cr0, cr1, rr0, rr1, ci0, ci1, ri0, ri1, p0, p1, phi0, phi1 = params

        c0, c1 = cr0 + 1j * ci0, cr1 + 1j * ci1

        cov0 = a_b_phi_2_cov(rr0, ri0, phi0)
        cov1 = a_b_phi_2_cov(rr1, ri1, phi1)

        eps = 1e-20
        return (
            -np.log(readout_distribution(s0, p0, c0, c1, cov0, cov1) +
                    eps).sum() -
            np.log(readout_distribution(s1, p1, c0, c1, cov0, cov1) +
                   eps).sum())

    res = minimize(loss, [
        s0.real.mean(),
        s1.real.mean(),
        s0.real.std(),
        s1.real.std(),
        s0.imag.mean(),
        s1.imag.mean(),
        s0.imag.std(),
        s1.imag.std(), 1, 0, 0, 0
    ],
                   args=(s0, s1),
                   bounds=[(None, None), (None, None), (1e-6, None),
                           (1e-6, None), (None, None), (None, None),
                           (1e-6, None), (1e-6, None), (0, 1), (0, 1),
                           (0, 2 * np.pi), (0, 2 * np.pi)])

    cr0, cr1, rr0, rr1, ci0, ci1, ri0, ri1, p0, p1, phi0, phi1 = res.x
    c0, c1 = cr0 + 1j * ci0, cr1 + 1j * ci1
    c0, c1 = c0 * scale + center, c1 * scale + center
    cov0 = a_b_phi_2_cov(rr0, ri0, phi0)
    cov1 = a_b_phi_2_cov(rr1, ri1, phi1)

    return (c0, c1, rr0 * scale, rr1 * scale, ri0 * scale, ri1 * scale, p0, p1,
            phi0, phi1, cov0 * scale**2, cov1 * scale**2)


def fit_readout_distribution2(s0, s1):
    center = 0.5 * (s0.mean() + s1.mean())
    s0, s1 = s0 - center, s1 - center

    scale = np.max([np.abs(s0.mean()), np.abs(s1.mean()), s0.std(), s1.std()])

    s0, s1 = s0 / scale, s1 / scale

    def loss(params, s0, s1):
        cr0, cr1, rr0, rr1, ci0, ci1, ri0, ri1, p0, p1 = params

        x0, y0 = cdf(None, s0.real)
        x1, y1 = cdf(None, s1.real)
        x2, y2 = cdf(None, s0.imag)
        x3, y3 = cdf(None, s1.imag)

        Y0 = mult_gaussian_cdf(x0, [cr0, cr1], [rr0, rr1], [p0, 1 - p0])
        Y1 = mult_gaussian_cdf(x1, [cr0, cr1], [rr0, rr1], [p1, 1 - p1])
        Y2 = mult_gaussian_cdf(x2, [ci0, ci1], [ri0, ri1], [p0, 1 - p0])
        Y3 = mult_gaussian_cdf(x3, [ci0, ci1], [ri0, ri1], [p1, 1 - p1])

        return (np.sum((Y0 - y0)**2) + np.sum((Y1 - y1)**2) + np.sum(
            (Y2 - y2)**2) + np.sum((Y3 - y3)**2))

    res = minimize(loss, [
        s0.real.mean(),
        s1.real.mean(),
        s0.real.std(),
        s1.real.std(),
        s0.imag.mean(),
        s1.imag.mean(),
        s0.imag.std(),
        s1.imag.std(), 1, 0
    ],
                   args=(s0, s1),
                   bounds=[(None, None), (None, None), (1e-6, None),
                           (1e-6, None), (None, None), (None, None),
                           (1e-6, None), (1e-6, None), (0, 1), (0, 1)])

    cr0, cr1, rr0, rr1, ci0, ci1, ri0, ri1, p0, p1 = res.x
    c0, c1 = cr0 + 1j * ci0, cr1 + 1j * ci1
    c0, c1 = c0 * scale + center, c1 * scale + center

    return (c0, c1, rr0 * scale, rr1 * scale, ri0 * scale, ri1 * scale, p0, p1,
            0, 0)


def get_threshold_info(s0, s1, thr=None, phi=None):
    from sklearn import svm

    s0, s1 = np.asarray(s0), np.asarray(s1)

    if phi is None:
        data = np.hstack([s0, s1])
        scale = 0.2 * np.abs(data).max()
        data /= scale
        target = np.hstack(
            [np.zeros_like(s0, dtype=float),
             np.ones_like(s1, dtype=float)])
        X = np.c_[np.real(data), np.imag(data)]
        clf = svm.LinearSVC()
        clf.fit(X, target)
        A, B, C = clf.coef_[0, 0], clf.coef_[0, 1], clf.intercept_[0]
        phi = np.arctan2(B, A)
        #thr = -scale * C / np.sqrt(A**2 + B**2)

    re0 = (s0 * np.exp(-1j * phi)).real
    re1 = (s1 * np.exp(-1j * phi)).real
    im0 = (s0 * np.exp(-1j * phi)).imag
    im1 = (s1 * np.exp(-1j * phi)).imag

    x = np.unique(np.hstack([re0, re1]))
    x.sort()
    a = cdf(x, re0)
    b = cdf(x, re1)
    c = a - b

    visibility = c.max()

    if thr is None:
        thr = x[c == visibility]
        thr = 0.5 * (thr.min() + thr.max())

    (c0, c1, rr0, rr1, ri0, ri1, p0, p1, phi0, phi1, cov0,
     cov1) = fit_readout_distribution(re0 + 1j * im0, re1 + 1j * im1)

    params_r = np.array([c0.real, c1.real, rr0, rr1, p0, p1, phi0])
    params_i = np.array([c0.imag, c1.imag, ri0, ri1, p0, p1, phi1])

    return {
        'threshold': thr,
        'phi': phi,
        'visibility': (visibility, cdf(thr, re0), cdf(thr, re1)),
        'signal': (re0, re1),
        'idle': (im0, im1),
        'center': (c0, c1),
        'params': (params_r, params_i),
        'std': (rr0, ri0, rr1, ri1, cov0, cov1),
        'cdf': (x, a, b, c)
    }


def getThresholdInfo(s0, s1):
    warnings.warn('getThresholdInfo is deprecated, use get_threshold_info',
                  DeprecationWarning, 2)
    return get_threshold_info(s0, s1)


def classifyData(data, measure_gates, avg=False):
    warnings.warn('classifyData is deprecated, use classify_data',
                  DeprecationWarning, 2)
    return classify_data(data, measure_gates, avg=avg)


def countState(state):
    warnings.warn('countState is deprecated, use count_state',
                  DeprecationWarning, 2)
    return count_state(state)



import secrets

from cryptography.exceptions import InvalidKey
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

SALT_SIZE = 32
KEY_SIZE = 32

__all__ = ['InvalidKey', 'encryptPassword', 'verifyPassword']


def encryptPassword(password):
    if isinstance(password, bytes):
        key_material = password
    else:
        key_material = str(password).encode('utf-8')

    salt = secrets.token_bytes(SALT_SIZE)
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                     length=KEY_SIZE,
                     salt=salt,
                     iterations=100000,
                     backend=default_backend())
    key = kdf.derive(key_material)
    return salt + key


def verifyPassword(password, hashed_password):
    if isinstance(password, bytes):
        key_material = password
    else:
        key_material = str(password).encode('utf-8')

    salt = hashed_password[:SALT_SIZE]
    key = hashed_password[SALT_SIZE:]
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(),
                     length=KEY_SIZE,
                     salt=salt,
                     iterations=100000,
                     backend=default_backend())
    kdf.verify(key_material, key)



import numpy as np
from waveforms.quantum.math import randomDensity, randomUnitary
from waveforms.quantum.tomo import (U_to_chi, applyChi, applyOp, dagger,
                                    pauli_basis, qpt, qptBases, qptInitList,
                                    qst_mle, qstOpList, tensorMatrix)


def applyChiDef(chi, rho, basis=pauli_basis):
    ret = np.zeros_like(rho, dtype=complex)
    N = int(round(np.log2(ret.shape[0])))
    for i, m in enumerate(qptBases(N, gates=basis)):
        for j, n in enumerate(qptBases(N, gates=basis)):
            ret += chi[i, j] * tensorMatrix(m) @ rho @ dagger(tensorMatrix(n))
    return ret


def test_qst():
    # QST
    N = 2  # 两个比特
    rho = randomDensity(2**N)

    # 模拟实验采集数据
    P = []
    for op in qstOpList(N):
        # 先做 op 变换，再测量 population
        z = np.diag(tensorMatrix(op) @ rho @ dagger(tensorMatrix(op)))
        P.append(z)

    rho_exp = qst_mle(P)
    assert np.allclose(rho, rho_exp, atol=0.01)


def test_qpt():
    # QPT
    riList = []
    rfList = []

    N = 2
    testGate = randomUnitary(4)

    for U in qptInitList(N):
        # 分别制备不同的初态
        ri = applyOp(U)
        # 做完操作后测末态的 QST
        rf = testGate @ ri @ dagger(testGate)
        riList.append(ri)
        rfList.append(rf)

    # 计算 chi 超算符
    chi = qpt(riList, rfList)

    assert np.allclose(chi, U_to_chi(testGate))

    rho_0 = randomDensity(testGate.shape[0])
    assert np.allclose(testGate @ rho_0 @ dagger(testGate),
                       applyChi(chi, rho_0))

    def checkChi(chi, U):
        rho_0 = randomDensity(U.shape[0])
        rho_F = applyChiDef(chi, rho_0)
        return np.allclose(rho_F, U @ rho_0 @ dagger(U))

    assert checkChi(chi, testGate)
    assert np.allclose(chi, U_to_chi(testGate))
    assert np.allclose(chi, qpt(rfList))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM binary operation expression."""

import warnings

from .node import Node


class BinaryOp(Node):
    """Node for an OPENQASM binary operation expression.
    children[0] is the operation, as a binary operator node.
    children[1] is the left expression.
    children[2] is the right expression.
    """

    def __init__(self, children):
        """Create the binaryop node."""
        super().__init__('binop', children, None)

    def qasm(self, prec=None, nested_scope=None):
        """Return the corresponding OPENQASM string."""
        if prec is not None:
            warnings.warn('Parameter \'BinaryOp.qasm(..., prec)\' is no longer '
                          'used and is being deprecated.', DeprecationWarning, 2)
        if nested_scope is not None:
            warnings.warn('Parameter \'BinaryOp.qasm(..., nested_scope)\' is no longer '
                          'used and is being deprecated.', DeprecationWarning, 2)
        return "(" + self.children[1].qasm() + self.children[0].value + \
               self.children[2].qasm() + ")"

    def latex(self, prec=None, nested_scope=None):
        """Return the corresponding math mode latex string."""
        if prec is not None:
            warnings.warn('Parameter \'BinaryOp.latex(..., prec)\' is no longer used '
                          'and is being deprecated.', DeprecationWarning, 2)
        if nested_scope is not None:
            warnings.warn('Parameter \'BinaryOp.latex(..., nested_scope)\' is no longer used '
                          'and is being deprecated.', DeprecationWarning, 2)
        try:
            from pylatexenc.latexencode import utf8tolatex
        except ImportError:
            raise ImportError("To export latex from qasm "
                              "pylatexenc needs to be installed. Run "
                              "'pip install pylatexenc' before using this "
                              "method.")
        return utf8tolatex(self.sym())

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        operation = self.children[0].operation()
        lhs = self.children[1].real(nested_scope)
        rhs = self.children[2].real(nested_scope)
        return operation(lhs, rhs)

    def sym(self, nested_scope=None):
        """Return the correspond symbolic number."""
        operation = self.children[0].operation()
        lhs = self.children[1].sym(nested_scope)
        rhs = self.children[2].sym(nested_scope)
        return operation(lhs, rhs)


# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM indexed id."""

import warnings

from .node import Node


class IndexedId(Node):
    """Node for an OPENQASM indexed id.
    children[0] is an id node.
    children[1] is an Int node.
    """

    def __init__(self, children):
        """Create the indexed id node."""
        super().__init__('indexed_id', children, None)
        self.id = children[0]  # pylint: disable=invalid-name
        self.name = self.id.name
        self.line = self.id.line
        self.file = self.id.file
        self.index = children[1].value

    def to_string(self, indent):
        """Print with indent."""
        ind = indent * ' '
        print(ind, 'indexed_id', self.name, self.index)

    def qasm(self, prec=None):
        """Return the corresponding OPENQASM string."""
        if prec is not None:
            warnings.warn('Parameter \'IndexedId.qasm(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        return self.name + "[%d]" % self.index
        


import numpy as np

from .fibheap import FibHeap, FibNode


def _generate_standrad_information(key, bayes_matrices):
    ret = []
    for i, k in enumerate(key):
        tmp = sorted([(p, j) for j, p in enumerate(bayes_matrices[i][k])],
                     key=lambda x: abs(x[0]),
                     reverse=True)
        ret.append((tuple(tmp), i))
    return tuple(ret)


def _generate_population(std_info, population):
    state, ret = [0] * len(std_info), population
    for i in range(len(std_info)):
        state[std_info[i][-1]] = std_info[i][0][0][1]
        ret *= std_info[i][0][0][0]
    return ret, tuple(state)


def _add_std_info_to_heap(state, population, heap, heap_nodes, max_n, min_p):
    if state in heap_nodes.keys():
        new_key = heap_nodes[state].key[0] * heap_nodes[state].key[
            -1] + population
        heap.change(heap_nodes[state],
                    (abs(new_key), state, (-1 if new_key < 0 else 1)))
    elif heap.key_number < max_n:
        heap_nodes[state] = FibNode(
            (abs(population), state, (-1 if population < 0 else 1)))
        heap.insert(heap_nodes[state])
    else:
        if abs(population) < max(min_p, heap.min_node.key[0]):
            return False
        trash = heap.extract_min()
        heap_nodes.pop(trash.key[1])
        heap_nodes[state] = FibNode(
            (abs(population), state, (-1 if population < 0 else 1)))
        heap.insert(heap_nodes[state])
    return True


def _update_count(population, std_info, heap, heap_nodes, max_n, min_p):
    queue = FibHeap()

    p, state = _generate_population(std_info, population)
    queue.insert(FibNode((-abs(p), state, std_info, 1 if p > 0 else -1)))

    for _ in range(max_n):
        node = queue.extract_min()
        if node is None or not _add_std_info_to_heap(
                node.key[1], -node.key[0] * node.key[-1], heap, heap_nodes,
                max_n, min_p):
            break

        tmp_info = sorted(list(node.key[2]),
                          reverse=True,
                          key=lambda x:
                          (len(x[0]), abs(x[0][1][0] / x[0][0][0])
                           if len(x[0]) > 1 else 0))

        for i in range(len(std_info)):
            if len(tmp_info[i][0]) == 1:
                break
            new_info = tuple([
                *tmp_info[:i], (tmp_info[i][0][1:], tmp_info[i][1]),
                *tmp_info[i + 1:]
            ])
            p, state = _generate_population(new_info, population)
            if heap.key_number < max_n or p > min(min_p, heap.min_node.key[0]):
                queue.insert(
                    FibNode((-abs(p), state, new_info, (-1 if p < 0 else 1))))
            elif abs(p) < max(min_p, heap.min_node.key[0]):
                break
    heap.consolidate()


def bayesian_correction_automatic_trimming(state,
                                           correction_matrices,
                                           size_lim=None,
                                           eps=None):
    from .fit.readout import count_state

    counts = count_state(state)
    shots = len(state)

    if size_lim is None:
        size_lim = shots * 4
    if eps is None:
        eps = 0.5 / size_lim

    heap_nodes = {}
    heap = FibHeap()

    for key, value in sorted(counts.items(), key=lambda x: x[1],
                             reverse=False):
        std_info = _generate_standrad_information(key, correction_matrices)
        _update_count(value / shots, std_info, heap, heap_nodes, size_lim, eps)

    result = {}
    while heap.key_number:
        node = heap.extract_min()
        if node is None:
            break
        result[node.key[1]] = node.key[0] * node.key[-1]
    return result


def bayesian_correction_in_subspace(state, correction_matrices, subspace):
    """Apply a correction matrix to a state.

    Args:
        state (np.array, dtype=int): The state to be corrected.
        correction_matrices (np.array): A list of correction matrices.
        subspace (np.array, dtype=int): The basis of subspace.

    Returns:
        np.array: The corrected state.

    Examples:
        >>> state = np.random.randint(2, size = (101, 1024, 4))
        >>> PgPe = np.array([[0.1, 0.8], [0.03, 0.91], [0.02, 0.87], [0.05, 0.9]])
        >>> correction_matrices = np.array(
            [np.array([[Pe, Pe - 1], [-Pg, 1 - Pg]]) / (Pe - Pg) for Pg, Pe in PgPe])
        >>> subspace = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [1, 1, 0, 0]])
        >>> result = bayesian_correction(state, correction_matrices, subspace)
        >>> result.shape
        (101, 5)
    """
    num_qubits = state.shape[-1]
    site_index = np.arange(num_qubits)

    shape = tuple([*state.shape[:-1], len(subspace)])
    state = state.reshape(-1, num_qubits)

    if len(subspace) < len(state):
        ret = []
        for target_state in subspace:
            A = np.prod(correction_matrices[site_index, target_state, state],
                        axis=-1)
            ret.append(A)
        ret = np.array(ret).T.reshape(shape)
    else:
        ret = []
        for bit_string in state:
            A = np.prod(correction_matrices[site_index, subspace, bit_string],
                        axis=-1)
            ret.append(A)
        ret = np.array(ret).reshape(shape)
    ret = ret.mean(axis=-2)
    return ret


def bayesian_correction(state,
                        correction_matrices,
                        *,
                        subspace=None,
                        size_lim=1024,
                        eps=1e-6):
    """Apply a correction matrix to a state.

    Args:
        state (np.array, dtype=int): The state to be corrected.
        correction_matrices (np.array): A list of correction matrices.
        subspace (np.array, dtype=int): The basis of subspace.
        size_lim (int): The maximum size of the heap.
        eps (float): The minimum probability of the state.

    Returns:
        np.array: The corrected state.

    Examples:
        >>> state = np.random.randint(2, size = (101, 1024, 4))
        >>> PgPe = np.array([[0.1, 0.8], [0.03, 0.91], [0.02, 0.87], [0.05, 0.9]])
        >>> correction_matrices = np.array(
            [np.array([[Pe, Pe - 1], [-Pg, 1 - Pg]]) / (Pe - Pg) for Pg, Pe in PgPe])
        >>> result = bayesian_correction(state, correction_matrices)
        >>> result.shape
        (101, 1024)
    """
    if subspace is None:
        state = np.array(state)
        if state.ndim == 2:
            return bayesian_correction_automatic_trimming(
                state, correction_matrices, size_lim, eps)
        shape = state.shape[:-2]
        state = state.reshape(-1, state.shape[-2], state.shape[-1])
        ret = []
        for s in state:
            ret.append(
                bayesian_correction_automatic_trimming(s, correction_matrices,
                                                       size_lim, eps))
        return np.array(ret).reshape(shape)
    else:
        return bayesian_correction_in_subspace(state, correction_matrices,
                                               subspace)



#from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup


def get_extensions():
    #from numpy.distutils.misc_util import get_numpy_include_dirs
    #from pathlib import Path

    extensions = [
        Extension('waveforms._waveform', ['src/waveform.c'],
                  include_dirs=['src']),
        Extension('waveforms.math._prime', ['src/prime.c'],
                  include_dirs=['src']),
        Extension('waveforms.sys.net._kcp', ['src/kcp.c', 'src/ikcp.c'],
                  include_dirs=['src']),
        # Extension(
        #     'waveforms.math.npufunc',
        #     ['src/multi_type_logit.c'],
        #     include_dirs=get_numpy_include_dirs(),
        #     library_dirs=[str(Path(p).parent / 'lib') for p in get_numpy_include_dirs()],
        #     libraries=["npymath"],
        #     # extra_compile_args=['-std=c99'],
        # ),
        # cythonize("waveforms/math/prime.pyx"),
    ]

    return extensions


setup(packages=find_packages(), ext_modules=get_extensions())



from waveforms.dicttree import *

step1 = {'a': {'x': 1, 'y': 2}, 'b': {'x': 3, 'y': 4}}
step2 = {'a': {'x': 2}, 'b': {'x': 3, 'y': 5}, 'c': {'x': 15}, 'd': 1, 'e': 1}
step3 = {'a': {'x': 3, 'y': 2, 'p': {'w': 10}}, 'c': {'x': 16}, 'd': 1, 'e': 2}
step4 = {'a': {'x': 4, 'y': 2}, 'b': {'x': 6, 'y': 4}, 'c': {'x': 17}}


def test_flattenDict():
    assert flattenDict({}) == {}
    assert flattenDict(step1) == {'a.x': 1, 'a.y': 2, 'b.x': 3, 'b.y': 4}


def test_foldDict():
    assert foldDict({}) == {}
    assert foldDict(flattenDict(step1)) == step1


def test_diff():
    assert diff({}, {}) == {}
    assert diff(step1, step1) == {}
    assert diff(step1, {}) == {'a': DELETE, 'b': DELETE}

    d1 = diff({}, step1)
    assert isinstance(d1['a'], Create) and d1['a'].n == {
        'x': 1,
        'y': 2
    } and d1['a'].replace == False
    assert isinstance(d1['b'], Create) and d1['b'].n == {
        'x': 3,
        'y': 4
    } and d1['b'].replace == False


def test_patch():
    assert patch(step1, {}) == step1
    assert patch(step1, diff(step1, {})) == {}
    assert patch({}, diff({}, step1)) == step1

    test = {
        'a': 1,
        'b': 2,
        'c': {
            'x': 3
        },
        'd': {
            'y': 4
        },
        'e': {
            'x': 1,
            'y': 4
        }
    }
    patch(test, {
        'a': Update(1, 10),
        'b': DELETE,
        'c': Create({'z': 12}, replace=True),
        'd': {
            'x': Create(5),
            'y': Update(4, 40)
        },
        'e': Create({
            'x': 2,
            'z': 13
        })
    },
          in_place=True)

    assert test == {
        'a': 10,
        'c': {
            'z': 12
        },
        'd': {
            'x': 5,
            'y': 40
        },
        'e': {
            'x': 2,
            'y': 4,
            'z': 13
        }
    }

    d1 = diff(step1, step2)
    d2 = diff(step2, step3)
    d3 = diff(step3, step4)

    assert patch(step1, d1) == step2
    assert patch(step2, d2) == step3
    assert patch(step3, d3) == step4


def test_merge():
    d1 = diff(step1, step2)
    d2 = diff(step2, step3)
    d3 = diff(step3, step4)

    assert merge(d1, d2, origin=step1) == diff(step1, step3)
    assert merge(d2, d3, origin=step2) == diff(step2, step4)

    assert patch(step1, merge(d1, d2)) == step3
    assert patch(step1, merge(merge(d1, d2), d3)) == step4
    assert patch(step1, merge(d1, merge(d2, d3))) == step4


def test_query():
    assert query_tree('a', step3) == {'x': 3, 'y': 2, 'p': {'w': 10}}
    assert query_tree('a.p', step3) == {'w': 10}
    assert query_tree('a.p.w', step3) == 10
    assert query_tree('a.p.w', step1) == (NOTSET, 'a.p')


def test_print():
    import io
    buf = io.StringIO()

    d1 = diff(step1, step3)
    print_diff(d1, file=buf)

    buf.seek(0)
    assert buf.read() == (
        "a.x                                      Update: 1 ==> 3\n"
        "a.p                                      Create: {'w': 10}\n"
        "c                                        Create: {'x': 16}\n"
        "d                                        Create: 1\n"
        "e                                        Create: 2\n"
        "b                                        Delete\n")



"""Multiple architecture support"""
from waveforms.qlisp.arch import get_arch, register_arch

from .base import (COMMAND, FREE, PUSH, READ, SYNC, TRIG, WRITE, Architecture,
                   general_architecture)



from __future__ import annotations

import ast
import functools
import hashlib
import operator
import pickle
import re
from datetime import datetime
from itertools import islice
from typing import Any, Generator, NamedTuple, Optional, Union

from sqlalchemy import (BigInteger, Boolean, Column, DateTime, ForeignKey,
                        Integer, LargeBinary, String, event, inspect)
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import aliased, backref, declarative_base, relationship
from sqlalchemy.orm.session import Session

from waveforms.dicttree import (DELETE, NOTSET, UNKNOW, Create, Singleton,
                                Update, foldDict, merge)


@compiles(BigInteger, 'sqlite')
def bi_c(element, compiler, **kw):
    return "INTEGER"


Base = declarative_base()


class KeyPattern(NamedTuple):
    like: bool
    fuzzy: bool
    pattern: str
    index: tuple
    end: bool
    case_insensitive: bool = False

    def format(self, matched: str) -> str:

        def f(slices):
            ret = []
            for s in slices:
                if isinstance(s, slice):
                    start, stop, step = s.start, s.stop, s.step
                    start = '' if start is None else start
                    stop = '' if stop is None else stop
                    step = '' if step is None else step
                    if s.step is None:
                        ret.append(f"{start}:{stop}")
                    else:
                        ret.append(f"{start}:{stop}:{step}")
                else:
                    ret.append(repr(s))
            s = ','.join(ret)
            return f"[{s}]"

        index_str = ''.join(f(slices) for slices in self.index)
        return matched + index_str

    @functools.lru_cache(maxsize=1)
    def re(self) -> re.Pattern:
        pattern = re.escape(self.pattern)

        if self.like:
            pattern = pattern.replace('%', '.*?')
            pattern = f"^{pattern}$"
        else:
            pattern = pattern.replace('%', '.*?')
            pattern = f"^((?!{pattern}).)+$"
        if self.case_insensitive:
            pattern = f"^((?i){pattern})$"
        return re.compile(pattern)


_key_pattern = re.compile(
    r'(?P<unlike>!?)(?P<pattern>[_A-Za-z\*][^\[\]\.]*)(?P<index>(\[(.*?)\])*)(?P<end>\.?)'
)


def _parse_key(key):
    ret = []

    for m in _key_pattern.finditer(key):
        index = []
        if m.group('index'):
            for slices_s in m.group('index')[1:-1].split(']['):
                slices = []
                for s in slices_s.split(','):
                    lst = s.split(':')
                    lst = [
                        None if l == '' else ast.literal_eval(l) for l in lst
                    ]
                    if 2 <= len(lst) <= 3:
                        expr = slice(*lst)
                    else:
                        expr = ast.literal_eval(s)
                    slices.append(expr)
                index.append(tuple(slices))
        pattern = m.group('pattern')
        index = tuple(index)
        like = m.group('unlike') == ''
        end = m.group('end') == ''
        if '*' in pattern:
            fuzzy = True
            pattern = pattern.replace('*', '%')
        else:
            fuzzy = False
        ret.append(KeyPattern(like, fuzzy, pattern, index, end))
    return ret


def _getitem(obj, index: tuple):
    for slices in index:
        if isinstance(slices, tuple) and len(slices) == 1:
            slices = slices[0]
        obj = operator.getitem(obj, slices)
    return obj


class _TREE(metaclass=Singleton):
    __slots__ = ()


TREE = _TREE()


class TreeRef():
    __slots__ = ('id')

    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return '{...}'


class RegistryValue(Base):
    __tablename__ = 'registry_values'

    id = Column(BigInteger, primary_key=True)
    hash = Column(LargeBinary(16), nullable=False, index=True)
    chunk = Column(LargeBinary)

    @property
    def value(self):
        return pickle.loads(self.chunk)

    @value.setter
    def value(self, value):
        self.set_value(value)

    def set_value(self, value):
        self.chunk = pickle.dumps(value, protocol=5)
        self.hash = hashlib.md5(self.chunk).digest()


class RegistryKey(Base):
    __tablename__ = 'registry_keys'

    left_id = Column(ForeignKey('registry_nodes.id'), primary_key=True)
    right_id = Column(ForeignKey('registry_nodes.id'), primary_key=True)
    key = Column(String(50), primary_key=True)

    value = relationship(
        "RegistryValue",
        secondary=
        "join(RegistryNode, RegistryValue, RegistryValue.id == RegistryNode.value_id)",
        primaryjoin="and_(RegistryKey.right_id == RegistryNode.id)",
        uselist=False,
        viewonly=True)

    def __repr__(self):
        return f'RegistryKey(key={self.key})'


class RegistryNode(Base):
    __tablename__ = 'registry_nodes'

    id = Column(BigInteger, primary_key=True)
    value_id = Column(ForeignKey('registry_values.id'), nullable=True)
    is_tree = Column(Boolean, default=False)

    value = relationship('RegistryValue')
    children = relationship("RegistryKey",
                            foreign_keys=[RegistryKey.left_id],
                            backref="parent")
    parents = relationship("RegistryKey",
                           foreign_keys=[RegistryKey.right_id],
                           backref="child")


class RegistrySnapshot(Base):
    __tablename__ = 'registry_snapshots'

    id = Column(BigInteger, primary_key=True)
    ctime = Column(DateTime, nullable=False, default=datetime.utcnow)
    root_id = Column(ForeignKey('registry_nodes.id'))
    previous_id = Column(BigInteger, ForeignKey('registry_snapshots.id'))

    followers = relationship("RegistrySnapshot",
                             backref=backref('previous', remote_side=[id]))

    root = relationship('RegistryNode')

    def __repr__(self):
        return f'RegistrySnapshot(id={self.id}, previous_id={self.previous_id})'


def create_tables(engine):
    Base.metadata.create_all(engine, checkfirst=True)


def _setitem(root: Union[RegistryNode, int], key: str,
             node: Union[RegistryNode, int]):
    edge = RegistryKey()
    edge.key = key
    if isinstance(root, RegistryNode):
        if root.id is None:
            edge.parent = root
        else:
            edge.left_id = root.id
    else:
        edge.left_id = root.id
    if isinstance(node, RegistryNode):
        if node.id is None:
            edge.child = node
        else:
            edge.right_id = node.id
    else:
        edge.right_id = node.id


def _create_value(session: Session, value: Any, cache: dict):
    value_obj = RegistryValue()
    value_obj.set_value(value)
    if value_obj.hash in cache:
        return cache[value_obj.hash]
    exist_value_id = session.query(RegistryValue.id).filter(
        RegistryValue.hash == value_obj.hash).one_or_none()
    if exist_value_id is None:
        return cache.setdefault(value_obj.hash, value_obj)
    else:
        return cache.setdefault(value_obj.hash, exist_value_id[0])


def _set_node_value(node: RegistryNode, value: Union[RegistryValue, int]):
    if isinstance(value, RegistryValue):
        if value.id is None:
            node.value = value
        else:
            node.value_id = value.id
    else:
        node.value_id = value


def _create_node(session: Session, value: dict,
                 value_cache: dict) -> RegistryNode:
    root = RegistryNode()
    if not isinstance(value, dict):
        root.is_tree = False
        if isinstance(value, (Create, Update)):
            value = value.n
        _set_node_value(root, _create_value(session, value, value_cache))
        return root

    root.is_tree = True
    _set_node_value(root, _create_value(session, TREE, value_cache))

    for k, v in value.items():
        if v is DELETE:
            continue
        elif isinstance(v, (Create, Update)):
            v = v.n
        _setitem(root, k, _create_node(session, v, value_cache))
    return root


def _patch(session: Session, root: Union[int, RegistryNode],
           diff: dict) -> RegistryNode:
    value_cache = {}
    if isinstance(root, int):
        root = session.get(RegistryNode, root)
    if not diff:
        return root

    if not isinstance(diff, dict):
        return _create_node(session, diff, value_cache)

    new_root = RegistryNode()
    new_root.is_tree = True
    _set_node_value(new_root, _create_value(session, TREE, value_cache))

    updated_keys = set()
    for edge in root.children:
        updated_keys.add(edge.key)
        if edge.key in diff:
            if diff[edge.key] is DELETE:
                continue
            elif isinstance(diff[edge.key], dict):
                node = _patch(session, edge.child, diff[edge.key])
            elif (isinstance(diff[edge.key], Create)
                  and not diff[edge.key].replace):
                node = _patch(session, edge.child, diff[edge.key].n)
            else:
                node = _create_node(session, diff[edge.key], value_cache)
        else:
            node = edge.child
        _setitem(new_root, edge.key, node)

    for k in set(diff.keys()) - updated_keys:
        if diff[k] is DELETE:
            continue
        elif isinstance(diff[k], (Create, Update)):
            _setitem(new_root, k, _create_node(session, diff[k].n,
                                               value_cache))
        else:
            _setitem(new_root, k, _create_node(session, diff[k], value_cache))

    return new_root


def _search_query(session, root_id, patterns):
    alias = [(k, aliased(RegistryKey)) for k in reversed(patterns)]

    node = aliased(RegistryNode)
    root = aliased(RegistryNode)

    path = tuple(a[1].key for a in reversed(alias))

    q = session.query(RegistryValue.chunk, node, *path)
    q = q.join(node, node.value_id == RegistryValue.id)
    right_side = None
    for k, a in alias:
        if right_side is None:
            join_codition = node.id == a.right_id
        else:
            join_codition = a.right_id == right_side.left_id
        q = q.join(a, join_codition)
        if k.pattern != '%':
            if k.case_insensitive:
                if k.like:
                    query_filter = a.key.ilike(k.pattern)
                else:
                    query_filter = a.key.not_ilike(k.pattern)
            else:
                if k.fuzzy:
                    if k.like:
                        query_filter = a.key.like(k.pattern)
                    else:
                        query_filter = a.key.not_like(k.pattern)
                else:
                    if k.like:
                        query_filter = a.key == k.pattern
                    else:
                        query_filter = a.key != k.pattern
            q = q.filter(query_filter)
        right_side = a
    q = q.join(root, root.id == right_side.left_id)
    if isinstance(root_id, int):
        q = q.filter(root.id == root_id)
    elif isinstance(root_id, (list, tuple, set)):
        q = q.filter(root.id.in_(root_id))
    else:
        raise TypeError('root_id must be int, list, tuple or set')
    #print(q.statement)
    return q


def search(session,
           root_id: int,
           patterns: list[KeyPattern],
           limit: Optional[int] = 100,
           offset: int = 0,
           depth: int = -1) -> dict[str, Any]:
    if depth == 0:
        return
    q = _search_query(session, root_id, patterns).offset(offset)
    if limit is not None and limit >= 0:
        q = q.limit(limit)
    for chunk, node, *path in q:
        if node.is_tree:
            value = export(session, node.id, depth=depth - 1)
        else:
            value = pickle.loads(chunk)
        try:
            value = _getitem(value, patterns[-1].index)
        except (IndexError, TypeError):
            continue
        yield '.'.join(k.format(s) for s, k in zip(path, patterns)), value


def export(session, root_id, depth=-1):
    if depth == 0:
        return TreeRef(root_id)

    pattern = KeyPattern(like=True,
                         fuzzy=True,
                         pattern='%',
                         index=(),
                         end=True)
    patterns = []
    ret = {}
    while True:
        depth -= 1
        patterns.append(pattern)
        sub_tree_count = 0
        for k, v in search(session,
                           root_id,
                           patterns=patterns,
                           limit=-1,
                           depth=1):
            if isinstance(v, TreeRef) and depth != 0:
                sub_tree_count += 1
                continue
            ret[k] = v
        if sub_tree_count == 0:
            break

    return foldDict(ret)


def _search_dict_iter(d, patterns, prefix):
    if len(patterns) == 0:
        yield '.'.join(prefix), d
    elif isinstance(d, dict):
        for k in d:
            if patterns[0].re().match(k):
                yield from _search_dict_iter(_getitem(d[k], patterns[0].index),
                                             patterns[1:],
                                             [*prefix, patterns[0].format(k)])


def search_dict(d, key, offset=0, limit=-1):
    patterns = _parse_key(key)
    if limit < 0:
        stop = None
    else:
        stop = offset + limit
    yield from islice(_search_dict_iter(d, patterns, []), offset, stop)


def search_flatten_dict(d, key, offset=0, limit=-1):
    patterns = _parse_key(key)
    pattern = re.compile('\.'.join([p.re().pattern[1:-1] for p in patterns]))
    for k, v in search_dict(d, key, offset, limit):
        if pattern.match(k):
            offset -= 1
            if offset < 0:
                yield k, v
            if limit == 0:
                return
            limit -= 1


class Registry():
    """A registry is a collection of snapshots.

    A registry is a collection of snapshots. Each snapshot is a tree of
    RegistryNode. Each RegistryNode has a value, which is a RegistryValue.
    RegistryValue can be a string, a number, a boolean, a list, a dict, or a
    reference to another RegistryNode.
    """

    def __init__(self,
                 session: Session,
                 snapshot: Union[None, int, dict, RegistrySnapshot] = None):
        """Create a registry object.

        Args:
            session (Session): SQLAlchemy session.
            snapshot (Union[None, int, dict, RegistrySnapshot], optional):
                Snapshot to use. Defaults to None.
        """
        self.session = session
        if snapshot is None:
            self.snapshot = session.query(RegistrySnapshot).order_by(
                RegistrySnapshot.id.desc()).first()
            if self.snapshot is None:
                self.init({})
        elif isinstance(snapshot, dict):
            self.init(snapshot)
        elif isinstance(snapshot, int):
            self.snapshot = session.get(RegistrySnapshot, snapshot)
        elif isinstance(snapshot, RegistrySnapshot):
            self.snapshot = snapshot
        else:
            raise ValueError('Invalid snapshot type')
        self.updates = {}

    def search(self,
               pattern: str,
               depth: int = 1,
               sub_tree: Union[TreeRef, dict, None] = None,
               limit: Optional[int] = None,
               offset: int = 0) -> Generator[tuple[str, Any], None, None]:
        """Search for keys matching pattern.

        Args:
            pattern: A string pattern to match keys. The pattern can contain
                wildcards. For example, 'a.b.*.d' will match 'a.b.c.d' and
                'a.b.x.d'. The pattern can also contain index. For example,
                'a.b[0].d' will match 'a.b[0].d' but not 'a.b[1].d'.
            depth: The depth of the tree to return. For example, if depth is 0,
                only the leaf nodes will be returned. If depth is 1, the leaf
                nodes and the nodes with one depth of children will be returned.
                If depth is -1, the entire tree will be returned.
            limit: The maximum number of keys to return.
            offset: The offset of the first key to return.

        Returns:
            A generator that yields a tuple of the key and the value.
        """
        patterns = _parse_key(pattern)
        if sub_tree is None:
            root_id = self.snapshot.root_id
        else:
            root_id = sub_tree.id
        yield from search(self.session,
                          root_id,
                          patterns,
                          limit=limit,
                          offset=offset,
                          depth=depth)

    def get(self, key, depth=-1):
        """Get the value of a key.

        Args:
            key: The key to get the value of.

        Returns:
            The value of the key.
        """
        if depth >= 0:
            depth += 1
        for k, v in self.search(key, limit=1, offset=0, depth=depth):
            return v
        else:
            return None

    def set(self, key, value, lazy: bool = False):
        """Set a value in the registry.

        Args:
            key: The key to set.
            value: The value to set.
            lazy: If True, the value will be set only when the registry is
                committed.
        """
        self.updates = merge(self.updates,
                             foldDict({key: Update(UNKNOW, value)}))
        if not lazy:
            self.commit()

    def delete(self, key, lazy: bool = False):
        """Delete a key from the registry.

        Args:
            key: The key to delete.
            lazy: If True, the key will be deleted only when the registry is
                committed.
        """
        self.updates = merge(self.updates, foldDict({key: DELETE}))
        if not lazy:
            self.commit()

    def update(self, updates: dict, lazy: bool = False):
        """Update the registry.

        Args:
            updates: A dictionary of updates. The keys are the keys to update
                and the values are the values to set. If the value is DELETE,
                the key will be deleted.
            lazy: If True, the updates will be applied only when the registry
                is committed.
        """
        self.updates = merge(self.updates, {
            key: Create(value, replace=False)
            for key, value in updates.items()
        })
        if not lazy:
            self.commit()

    def commit(self):
        """Commit the updates to the registry."""
        root = _patch(self.session, self.snapshot.root_id, self.updates)
        snapshot = RegistrySnapshot(previous_id=self.snapshot.id)
        snapshot.root = root
        self.session.add(snapshot)
        try:
            self.session.commit()
        except:
            self.session.rollback()
            raise
        self.snapshot = snapshot
        self.updates.clear()

    def rollback(self):
        """Rollback the updates."""
        self.updates.clear()

    def init(self,
             value: Union[dict, RegistryNode] = None) -> RegistrySnapshot:
        value_cache = {}
        snapshot = RegistrySnapshot()
        if value is None:
            value = {}
        if isinstance(value, RegistryNode):
            snapshot.root = value
        else:
            snapshot.root = _create_node(self.session, value, value_cache)
        self.session.add(snapshot)
        try:
            self.session.commit()
        except:
            self.session.rollback()
            raise
        self.snapshot = snapshot

    def export(self, sub_tree: TreeRef = None, depth: int = -1):
        """Export the registry to a dictionary.

        Args:
            sub_tree: The sub-tree to export.
            depth: The depth of the tree to return. For example, if depth is 0,
                only the leaf nodes will be returned. If depth is 1, the leaf
                nodes and the nodes with one depth of children will be returned.
                If depth is -1, the entire tree will be returned.
        """
        if sub_tree is None:
            root_id = self.snapshot.root_id
        else:
            root_id = sub_tree.id
        return export(self.session, root_id, depth)

    def previous(self):
        """Get the previous snapshot."""
        return Registry(self.session, self.snapshot.previous)

    @staticmethod
    def from_file(path: str, snapshot: Optional[int] = None) -> Registry:
        """Create a registry from a file.

        Args:
            path: The path to the file.
            snapshot: The snapshot to use. If None, the latest snapshot will be
                used.
        """
        url = f'sqlite:///{path}'
        return Registry.from_url(url, snapshot)

    @staticmethod
    def from_url(url: str, snapshot: Optional[int] = None) -> Registry:
        """Create a registry from a URL.

        Args:
            url: The URL to the database.
            snapshot: The snapshot to use. If None, the latest snapshot will be
                used.
        """
        from sqlalchemy import create_engine
        from sqlalchemy.orm import sessionmaker

        engine = create_engine(url)
        Base.metadata.create_all(engine, checkfirst=True)
        Session = sessionmaker(bind=engine)
        session = Session()
        return Registry(session, snapshot)

    def __del__(self):
        self.session.close()



from collections import defaultdict

from numpy import mod, pi

from .base import QLispError, gateName


def _lookup(name, env):
    try:
        return env.get(name, name)
    except:
        return name


def lookup(st, env):
    if isinstance(st[1], tuple):
        return (st[0], tuple(_lookup(q, env) for q in st[1]))
    elif isinstance(st[1], (str, int)):
        return (st[0], _lookup(st[1], env))
    else:
        return st


def define_macro(name, value, env):
    env[name] = value


def call_macro(gate, st):
    qubits = st[1]
    if isinstance(st[0], str):
        args = ()
    else:
        args = st[0][1:]
    try:
        yield from gate(qubits, *args)
    except:
        raise QLispError(f'extend macro {st} error.')


def extend_control_gate(st, scope):
    # TODO
    gate, qubits = st
    if isinstance(gate[1], str):
        if gate[1] == 'Z':
            return [('CZ', qubits)]
        elif gate[1] == 'X':
            return [('Cnot', qubits)]
        else:
            return [st]
    else:
        return [st]


def extend_macro(qlisp, lib, env=None):
    if env is None:
        env = {}
    for st in qlisp:
        if gateName(st) == 'define':
            define_macro(st[1], st[2], env)
        elif gateName(st) == 'C':
            st = lookup(st, env)
            yield from extend_control_gate(st, lib)
        else:
            st = lookup(st, env)
            gate = lib.getGate(gateName(st))
            if gate is None:
                yield st
            else:
                for st in call_macro(gate, st):
                    yield from extend_macro([st], lib, env)


_VZ_rules = {}


def add_VZ_rule(gateName, rule):
    _VZ_rules[gateName] = rule


def remove_VZ_rule(gateName, rule):
    del _VZ_rules[gateName]


def _VZ_P(st, phaseList):
    return [], [mod(phaseList[0] + st[0][1], 2 * pi)]


def _VZ_rfUnitary(st, phaseList):
    (_, theta, phi), qubit = st
    return [(('rfUnitary', theta, phi - phaseList[0]), qubit)], phaseList


def _VZ_clear(st, phaseList):
    return [st], [0] * len(phaseList)


def _VZ_exchangable(st, phaseList):
    return [st], phaseList


def _VZ_swap(st, phaseList):
    return [st], phaseList[::-1]


add_VZ_rule('P', _VZ_P)
add_VZ_rule('rfUnitary', _VZ_rfUnitary)
add_VZ_rule('Reset', _VZ_clear)
add_VZ_rule('Measure', _VZ_clear)
add_VZ_rule('CZ', _VZ_exchangable)
add_VZ_rule('I', _VZ_exchangable)
add_VZ_rule('Barrier', _VZ_exchangable)
add_VZ_rule('Delay', _VZ_exchangable)
add_VZ_rule('iSWAP', _VZ_swap)
add_VZ_rule('SWAP', _VZ_swap)


def exchangeRzWithGate(st, phaseList, lib):
    gate = gateName(st)
    if gate in _VZ_rules:
        return _VZ_rules[gate](st, phaseList)
    else:
        raise Exception('Unknow VZ exchange rule.')


def reduceVirtualZ(qlisp, lib):
    hold = defaultdict(lambda: 0)

    for st in qlisp:
        target = st[1]
        if isinstance(target, (int, str)):
            target = (target, )
        try:
            stList, phaseList = exchangeRzWithGate(st,
                                                   [hold[q] for q in target],
                                                   lib)
            yield from stList
            for q, p in zip(target, phaseList):
                hold[q] = mod(p, 2 * pi)
        except:
            for q in target:
                if hold[q] != 0:
                    yield (('P', hold[q]), q)
                    hold[q] = 0
            yield st

    for q in hold:
        if hold[q] != 0:
            yield (('P', hold[q]), q)



import bisect
import functools
import itertools
import random
from math import ceil, floor, inf, isqrt, log
from typing import Generator, Iterator, List, Optional, Set, Union

SIEVE_LIMIT = 50000


def __sieve(lst: List[int], p_lst: List[int]) -> None:
    """Sieve of Eratosthenes"""
    p = next(lst)
    p_lst.append(p)
    for num in lst:
        if all(num % p != 0
               for p in itertools.takewhile(lambda x, lim=isqrt(num): x <= lim,
                                            p_lst)):
            p_lst.append(num)


__least_primes: List[int] = [2]

__sieve(iter(range(3, SIEVE_LIMIT, 2)), __least_primes)

__primes_set: Set[int] = set(__least_primes)


def _MillerRabin(n: int, a: int) -> bool:
    """
    Miller-Rabin test with base a

    Args:
        n (int): number to test
        a (int): base

    Returns:
        bool: result
    """
    d = n - 1
    while (d & 1) == 0:
        d >>= 1
    t = pow(a, d, n)
    while d != n - 1 and t != n - 1 and t != 1:
        t = (t * t) % n
        d <<= 1
    return t == n - 1 or (d & 1) == 1


def millerRabinTest(q: int) -> bool:
    if q < 1373653:
        return all(_MillerRabin(q, a) for a in [2, 3])
    elif q < 9080191:
        return all(_MillerRabin(q, a) for a in [31, 73])
    elif q < 4759123141:
        return all(_MillerRabin(q, a) for a in [2, 7, 61])
    elif q < 2152302898747:
        return all(_MillerRabin(q, a) for a in [2, 3, 5, 7, 11])
    elif q < 3474749660383:
        return all(_MillerRabin(q, a) for a in [2, 3, 5, 7, 11, 13])
    elif q < 341550071728321:
        return all(_MillerRabin(q, a) for a in [2, 3, 5, 7, 11, 13, 17])
    elif q < 3825123056546413051:
        return all(
            _MillerRabin(q, a) for a in [2, 3, 5, 7, 11, 13, 17, 19, 23])
    elif q < 318665857834031151167461:
        return all(
            _MillerRabin(q, a)
            for a in [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37])
    elif q < 3317044064679887385961981:
        return all(
            _MillerRabin(q, a)
            for a in [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41])
    else:
        bases = random.sample(__primes_set, 20)
        return all(_MillerRabin(q, a) for a in bases)


def is_prime(q: int) -> bool:
    if q in __primes_set:
        return True
    if q <= SIEVE_LIMIT or q % 2 == 0 or q % 3 == 0:
        return False
    if millerRabinTest(q):
        __primes_set.add(q)
        return True
    else:
        return False


class _Primes:

    @staticmethod
    def greater_than(x: int) -> Iterator[int]:
        """
        generate primes greater than `x`

        Args:
            x (int): lower boundary

        Yields:
            Iterator[int]: primes
        """
        if x <= 2:
            yield 2
            q = 3
        elif x % 2 == 0:
            q = x + 1
        else:
            q = x
        while True:
            if is_prime(q):
                yield q
            q += 2

    @staticmethod
    def less_than(x: int) -> Iterator[int]:
        """
        generate primes less than x

        Args:
            x (int): upper boundary

        Yields:
            Iterator[int]: primes
        """
        if x <= 2:
            return
        if x == 3:
            yield 2
            return
        if x % 2 == 0:
            q = x - 1
        else:
            q = x - 2
        while True:
            if is_prime(q):
                yield q
            q -= 2
            if q == 1:
                yield 2
                break

    @staticmethod
    def take(count: int) -> Iterator[int]:
        """
        take first `count` primes
        """
        for _, q in zip(range(count), Primes.greater_than(1)):
            yield q

    @staticmethod
    def pi(x):
        """gives the number of primes less than or equal to x."""
        return primePi(x)

    def __iter__(self) -> Iterator[int]:
        yield from self.greater_than(1)

    def __contains__(self, num: int) -> bool:
        return is_prime(num)

    def __getitem__(self, n: Union[int, slice]) -> Union[int, list]:
        if isinstance(n, int):
            return prime(n + 1)
        elif isinstance(n, slice):
            start, stop, step = n.start, n.stop, n.step
            if start is None:
                start = 0
            if step is None:
                step = 1
            if stop is None:
                stop = inf
            return (prime(i + 1) for i in itertools.takewhile(
                lambda n: n < stop, itertools.count(start, step)))
        elif isinstance(n, tuple):
            return [prime(i + 1) for i in n]
        else:
            raise TypeError(
                f'indices must be integers or slices, not {type(n)}')

    def __call__(self):
        return self

    def __and__(self, other):
        if isinstance(other, (Generator, range)):
            return filter(is_prime, other)
        return {x for x in set(other) if is_prime(x)}

    def __rand__(self, other):
        return self.__and__(other)


def next_prime(x: int) -> int:
    """下一个质数"""
    if x < __least_primes[-1]:
        index = bisect.bisect(__least_primes, x)
        return __least_primes[index]
    if x % 2 == 0:
        x += 1
    else:
        x += 2
    while True:
        if is_prime(x):
            return x
        else:
            x += 2


def previous_prime(x: int) -> Optional[int]:
    """前一个质数"""
    if x <= 2:
        return None
    if x == 3:
        return 2
    if x % 2 == 0:
        x -= 1
    else:
        x -= 2
    while True:
        if is_prime(x):
            return x
        else:
            x -= 2


@functools.lru_cache(maxsize=None)
def _prime(n: int) -> int:
    a = floor(n * (log(n) + log(log(n)) - 1 + (log(log(n)) - 2.1) / log(n)))
    if a % 2 == 0:
        a += 1
    count = primePi(a)
    if count == n and is_prime(a):
        return a
    while True:
        a = next_prime(a)
        count += 1
        if count == n:
            return a


def prime(n: int) -> int:
    """第 n 个质数"""
    if n <= 0:
        raise ValueError('n must be positive')
    elif n <= len(__least_primes):
        return __least_primes[n - 1]

    return _prime(n)


def _Phi(m: int, b: int) -> int:
    if b == 0:
        return m
    elif b == 1:
        return m - m // 2
    elif m == 0:
        return 0
    elif m in [1, 2, 3, 4]:
        return 1
    else:
        return _cached_Phi(m, b)


@functools.lru_cache(maxsize=None)
def _cached_Phi(m: int, b: int) -> int:
    ret = m
    for i in range(b):
        ret -= _Phi(m // prime(i + 1), i)
    return ret


@functools.lru_cache(maxsize=None)
def _primePi(m: int) -> int:
    y = ceil(m**(1 / 3))
    n = primePi(y)

    P2 = 0
    if y % 2 == 0:
        q = y + 1
    else:
        q = y + 2
    while q <= isqrt(m):
        if is_prime(q):
            P2 += primePi(m // q) - primePi(q) + 1
        q += 2

    return _Phi(m, n) + n - 1 - P2


def primePi(m: int) -> int:
    if m <= __least_primes[-1]:
        return bisect.bisect(__least_primes, m)

    return _primePi(m)


Primes = _Primes()

__all__ = ['Primes', 'prime', 'is_prime', 'next_prime', 'primePi']



from ..qlisp.simulator import applySeq, regesterGateMatrix, seq2mat
from . import xeb
from .circuit.qlisp import compile, libraries
from .circuit.qlisp import std as stdlib
from .circuit.qlisp.config import Config, getConfig, setConfig
from .circuit.qlisp.qlisp import QLispError
from .circuit.qlisp.utils import (ALLXY, CP, CPMG, DD, UDD, XY4, XY8, XY16,
                                  Ramsey, SpinEcho)
from .tomo import qpt, qptInitList, qst, qst_mle, qstOpList



from waveforms.qlisp.macro import (_VZ_P, _VZ_clear, _VZ_exchangable,
                                   _VZ_rfUnitary, _VZ_swap, add_VZ_rule,
                                   exchangeRzWithGate, extend_macro,
                                   reduceVirtualZ, remove_VZ_rule)



import numpy as np
import scipy.constants as const


def Svv(f, T, Z=lambda f: 50 * np.ones_like(f)):
    """
    power spectral density of the series noise voltage

    Johnson-Nyquist noise
    https://en.wikipedia.org/wiki/Johnson%E2%80%93Nyquist_noise
    
    f : list of frequency
    T : temperature
    Z : frequency-dependent complex electrical impedance
        (default 50 Ohm)
    """
    from scipy.constants import h, k

    if callable(Z):
        R = np.real(Z(f))
    else:
        R = np.real(Z)
    x = h * f / (k * T)
    x, R, f, T = np.broadcast_arrays(x, R, f, T)
    ret = np.zeros_like(x)
    mask1 = x < 37
    mask2 = x >= 37
    ret[mask1] = 4 * k * T[mask1] * R[mask1] * x[mask1] / (np.exp(x[mask1]) -
                                                           1)
    ret[mask2] = 4 * h * f[mask2] * R[mask2] * np.exp(-x[mask2])
    return ret


def Sii(f, T, Z=lambda f: 50 * np.ones_like(f)):
    """
    power spectral density of the series noise voltage

    Johnson-Nyquist noise
    https://en.wikipedia.org/wiki/Johnson%E2%80%93Nyquist_noise
    
    f : list of frequency
    T : temperature
    Z : frequency-dependent complex electrical impedance
        (default 50 Ohm)
    """
    from scipy.constants import h, k

    if callable(Z):
        Y = np.real(1 / Z(f))
    else:
        Y = np.real(1 / Z)
    x = h * f / (k * T)
    x, Y, f, T = np.broadcast_arrays(x, Y, f, T)
    ret = np.zeros_like(x)
    mask1 = x < 37
    mask2 = x >= 37
    ret[mask1] = 4 * k * T[mask1] * Y[mask1] * x[mask1] / (np.exp(x[mask1]) -
                                                           1)
    ret[mask2] = 4 * h * f[mask2] * Y[mask2] * np.exp(-x[mask2])
    return ret


def atts(f, atts=[], input=None):
    """
    power spectral density at MXC
    
    f : list of frequency
    atts: list of tuples (temperature, attenuator)    
    """
    if input is not None:
        spec = input
    else:
        spec = 0.5 * Svv(f, 300)
    for T, att in atts:
        A = 10**(-att / 10)
        spec = spec / A + 0.5 * Svv(f, T) * (A - 1) / A
    return spec


def atts_and_heat(f, atts=[], input=None):
    """
    power spectral density at MXC
    
    f : list of frequency
    atts: list of tuples (temperature, attenuator)    
    """
    heat = np.zeros_like(f)
    if input is not None:
        spec = input
    else:
        spec = 0.5 * Svv(f, 300)
    for T, att in atts:
        A = 10**(-att / 10)
        heat += 300 / T * (A - 1) / A * spec
        spec = spec / A + 0.5 * Svv(f, T) * (A - 1) / A
    return spec, heat


def thermal_excitation(T, f01, *levels):
    """
    p1 / p0 = exp(-beta*E01)
    p0 + p1 = 1
    p1 = tanh(-beta * E01 / 2) / 2 + 1 / 2
    """
    if len(levels) == 0:
        return 0.5 * np.tanh(-0.5 * const.h * f01 / const.k / T) + 0.5
    else:
        levels = np.hstack([[f01], levels])
        pp = np.exp(-const.h * levels / const.k / T)
        p0 = 1 / (np.sum(pp) + 1)
        return pp * p0


def Z_in(w, ZL, l, Z0=50, v=1e8):
    """Impedance of the transmission line
    """
    a = 1j * np.tan(w * l / v)
    #return Z0*(np.tanh(1j*w*l/v)+np.arctanh(ZL/Z0))
    return Z0 * (ZL + Z0 * a) / (Z0 + ZL * a)


def S21(w, l, ZL, Z0=50, v=1e8):
    """S21 of the transmission line
    """
    z = Z_in(w, ZL, l, Z0, v) / Z0
    phi = w * l / v
    #return (1+Z0/z)*np.exp(-1j*w*l/v)/2+(1-Z0/z)*np.exp(1j*w*l/v)/2
    return np.cos(phi) - 1j / z * np.sin(phi)


def lorentzianAmp(x, x0, gamma):
    """lorentzian peak"""
    return 1 / (1 + ((x - x0) / gamma)**2)


def lorentzian(x, x0, gamma):
    """complex lorentzian peak
    
    lorentzianAmp(x, x0, gamma) = lorentzian(x, x0, gamma) * conj(lorentzian(x, x0, gamma))
    """
    return 1 / (1 + 1j * (x - x0) / gamma)


def gaussian(x, x0, sigma):
    """gaussian peak"""
    return np.exp(-0.5 * ((x - x0) / sigma)**2)


def lorentzianGaussian(x, x0, gamma, sigma):
    """complex lorentzian peak
    """
    if gamma == 0:
        return gaussian(x, x0, sigma)
    elif sigma == 0:
        return lorentzian(x, x0, gamma)
    else:
        t = np.arange(-3 * sigma, 3 * sigma, x[1] - x[0])
        t -= t.mean()
        ker = gaussian(t, 0, sigma)
        ker /= ker.sum()
        return np.convolve(lorentzian(x, x0, gamma), ker, mode='same')


def lorentzianSpace(f0, gamma, numOfPoints):
    """frequencies for lorentzian peak

    Args:
        f0 : center frequency
        gamma : width
        numOfPoints : number of points
    
    Returns:
        frequencies: np.array
    """
    theta = np.linspace(-np.pi, np.pi, numOfPoints + 2)[1:-1]
    return f0 - gamma * np.tan(theta)


def peaks(x, peaks, background=0):
    """
    peaks: list of (center, width, amp, shape)
           shape should be either 'gaussian' or 'lorentzian'
    background: a float, complex or ndarray with the same shape of `x`
    """
    ret = np.zeros_like(x)
    for center, width, amp, shape in peaks:
        if shape == 'gaussian':
            ret += amp * gaussian(x, center, width)
        elif shape == 'lorentzian':
            ret += amp * lorentzian(x, center, width)
        else:
            ret += amp * lorentzian(x, center, width)

    return ret + background


def complexPeaks(x, peaks, background=0):
    """
    peaks: list of (center, width, amp)
    background: a float, complex or ndarray with the same shape of `x`
    """
    ret = np.zeros_like(x, dtype=np.complex)
    for x0, gamma, A, *_ in peaks:
        ret += A * lorentzian(x, x0, gamma)
    return ret + background


def decay(t, tau):
    """
    exponential decay
    """
    a = -(1 / np.asarray(tau))**(np.arange(len(tau)) + 1)
    a = np.hstack([a[::-1], [0]])
    return np.exp(np.poly1d(a)(t))


def oscillation(t, spec=((1, 1), ), amplitude=1, offset=0):
    """
    oscillation
    """
    ret = np.zeros_like(t, dtype=np.complex)
    for A, f in spec:
        ret += A * np.exp(2j * np.pi * f * t)
    return amplitude * np.real(ret) + offset


def correlation(in1, in2, sample_rate=1, sample_rate2=None):
    from math import lcm

    from scipy.signal import correlate, correlation_lags, resample

    sample_rate1 = round(sample_rate)
    if sample_rate2 is None:
        sample_rate2 = round(sample_rate)
    else:
        sample_rate2 = round(sample_rate2)
    sample_rate = lcm(sample_rate1, sample_rate2)

    in1_len = len(in1) * sample_rate // sample_rate1
    in2_len = len(in2) * sample_rate // sample_rate2
    if in1_len > len(in1):
        in1 = resample(in1, in1_len)
    if in2_len > len(in2):
        in2 = resample(in2, in2_len)

    x = correlation_lags(in1_len, in2_len, mode='full') / float(sample_rate)
    y = correlate(in1, in2, mode='full', method='fft') / np.sqrt(
        correlate(in1, in1, mode='valid')[0] *
        correlate(in2, in2, mode='valid')[0])

    return x, y


def median_filter(x, window=5):
    """
    median filter

    Parameters
    ----------
    x : array_like
        input array
    window : int
        half window size

    Returns
    -------
    y : ndarray
        filtered array
    """
    y = []
    for i in range(len(x)):
        if i - window < 0:
            y.append(np.median(x[0:i + window + 1]))
        elif i + window > len(x):
            y.append(np.median(x[i - window:i + window + 1]))
        else:
            y.append(np.median(x[i - window:i + window + 1]))
    return np.array(y)



from functools import reduce
from itertools import chain, product
from typing import Optional, Union

import numpy as np
import scipy.sparse as sp
from scipy.linalg import eigh, expm, logm, sqrtm
from waveforms.math import (fit_circle, fit_cross_point, fit_pole, getFTMatrix,
                            lin_fit)
from waveforms.math.signal import decay, oscillation
from waveforms.qlisp.simulator.mat import (BellPhiM, BellPhiP, BellPsiM,
                                           BellPsiP, U, Unitary2Angles, fSim,
                                           issparse, make_immutable, rfUnitary,
                                           sigmaI, sigmaX, sigmaY, sigmaZ)

# Paulis
s0, s1, s2, s3 = sigmaI(), sigmaX(), sigmaY(), sigmaZ()

# Bell states
phiplus, phiminus = BellPhiP(), BellPhiM()
psiplus, psiminus = BellPsiP(), BellPsiM()


def tensor(matrixes: np.ndarray) -> np.ndarray:
    """Compute the tensor product of a list (or array) of matrixes"""
    return reduce(np.kron, matrixes)


def sigma(j: int, N: int = 1) -> np.ndarray:
    """
    """
    s = [s0, s1, s2, s3]
    dims = [4] * N
    idx = np.unravel_index(j, dims)
    return tensor(s[x] for x in idx)


def basis(x: Union[int, str], dim: int = 2) -> np.ndarray:
    '''
    Returns a single element of a state vector.  Each component is
    either a digit (0,1,2,...) representing the computational basis,
    or one or two letters representing a special state (H,V,D,A,R,L,
    X,Y,Z,Xm,Ym,Zm.  B1..4 are the 4 bell states
    '''
    d = {
        'H': [1, 0],
        'V': [0, 1],
        'D': [1, 1],
        'A': [1, -1],
        'R': [1, 1j],
        'L': [1, -1j],
        'X': [1, 1],
        'Xp': [1, 1],
        'Xm': [1, -1],
        'Y': [1, 1j],
        'Yp': [1, 1j],
        'Ym': [1, -1j],
        'Z': [1, 0],
        'Zp': [1, 0],
        'Zm': [0, 1],
        'B1': phiminus,
        'B2': phiplus,
        'B3': psiminus,
        'B4': psiplus
    }
    if isinstance(x, int) or x.isdigit():
        rv = np.zeros((dim), dtype=complex)
        rv[int(x)] = 1
    else:
        rv = np.array(d[x], dtype=complex)
        rv = rv / np.linalg.norm(rv)
    return rv


def commutator(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """Compute the commutator of two matrices"""
    return A @ B - B @ A


def dagger(vector: np.ndarray) -> np.ndarray:
    """Compute the hermitian conjugate of a vector or matrix"""
    return vector.conjugate().transpose()


def normalize(X: np.ndarray) -> np.ndarray:
    """Normalize the representation of X.  This means:
    For state vectors, X' * X = 1 and the first non-zero
    element of X is positive real.  For density matricies,
    X is hermetian, unit trace, and non-negative"""
    if X.ndim == 1:
        nz = X[np.nonzero(X)][0]
        X *= nz.conj()
        amp = np.abs(dagger(X) @ X)
        X /= np.sqrt(amp)
    else:
        (d, U) = eigh(X)
        d = d.real
        d[d < 0] = 0
        X = U @ np.diag(d) @ dagger(U)
        X = dagger(X) + X
        X /= np.trace(X)
    return X


def psi2rho(psi: np.ndarray) -> np.ndarray:
    '''
    Converts input state vector to density matrix represenation.  If psi
    is already a density matrix, leave it alone, so this function can be
    used to coerce input arguments to the desired type.
    '''
    if psi.ndim == 1:
        return normalize(np.outer(psi, psi.conj()))
    else:  # Already a density matrix
        return normalize(psi)


def rho2v(rho):
    """密度矩阵转相干矢
    """
    N = rho.shape[0]
    indices = np.tril_indices(N, -1)
    z = np.diag(rho).real
    Z = np.cumsum(z[:-1]) - z[1:] * np.arange(1, N)
    return np.hstack((rho[indices].real, rho[indices].imag, Z))


def v2rho(V):
    """相干矢转密度矩阵"""
    N = int(np.sqrt(len(V) + 1))
    assert N**2 - 1 == len(V)

    X = V[:(N**2 - N) // 2]
    Y = V[(N**2 - N) // 2:(N**2 - N)]
    Z = V[(N**2 - N):]

    A = np.hstack([Z[:0:-1] - Z[-2::-1], Z[0]])
    zn = (1 - A.sum()) / N
    A = A / np.arange(N - 1, 0, -1)
    z = np.cumsum(np.hstack([zn, A]))[::-1]

    rho = np.diag(z / 2).astype(complex)

    rho[np.tril_indices(N, -1)] = X + 1j * Y
    rho += dagger(rho)

    return normalize(rho)


def Hermitian2v(H):
    N = H.shape[0]
    indices = np.triu_indices(N, 1)
    return np.hstack((H[indices].real, H[indices].imag, np.diag(H).real))


def v2Hermitian(V):
    N = int(round(np.sqrt(len(V))))

    X = V[:(N**2 - N) // 2]
    Y = V[(N**2 - N) // 2:(N**2 - N)]
    Z = V[(N**2 - N):]
    H = np.diag(Z / 2).astype(np.complex)

    H[np.triu_indices(N, 1)] = X + 1j * Y
    H += dagger(H)
    return H


def unitary2v(U):
    H = -1j * logm(U)
    return Hermitian2v(H)


def v2unitary(V):
    H = v2Hermitian(V)
    return expm(1j * H)


def rho2bloch(rho):
    """
    与 rho2v 类似，只不过是选择以 sigma_i \\otimes sigma_j ... 为基底来展开,
    只适用于N比特态,即 rho 的形状只能是 (2 ** N, 2 ** N)
    """
    bases = [s0, s1, s2, s3]
    N = int(np.log2(rho.shape[0]))
    return np.asarray([
        np.sum(rho * reduce(np.kron, s).T).real
        for s in product(bases, repeat=N)
    ][1:])


def bloch2rho(V):
    """
    rho2bloch 的逆函数
    """
    bases = [s0, s1, s2, s3]
    N = int(np.log2(len(V) + 1) / 2)
    rho = reduce(np.add,
                 (a * reduce(np.kron, s)
                  for (a, s) in zip(chain([1], V), product(bases, repeat=N))))
    return rho / rho.shape[0]


def traceDistance(rho: np.ndarray, sigma: np.ndarray) -> float:
    """Compute the trace distance between matrixes rho and sigma
    See Nielsen and Chuang, p. 403
    """
    A = rho - sigma
    return np.real(np.trace(sqrtm(dagger(A) @ A))) / 2.0


def fidelity(rho: np.ndarray, sigma: np.ndarray) -> float:
    '''
    The fidelity of two quantum states.
    '''
    rho = psi2rho(rho)
    sigma = psi2rho(sigma)
    rhosqrt = sqrtm(rho)
    return np.real(np.trace(sqrtm(rhosqrt @ sigma @ rhosqrt)))


def entropy(rho: np.ndarray) -> float:
    '''von Neumann / Shannon entropy function'''
    if rho.ndim == 1:  # pure states have no entropy
        return 0
    v = np.linalg.eigvalsh(rho).real
    return -np.sum(v[v > 0] * np.log2(v[v > 0]))


def concurrence(rho: np.ndarray) -> float:
    """Concurrence of a two-qubit density matrix.
    see http://qwiki.stanford.edu/wiki/Entanglement_of_Formation
    """
    yy = np.array([[ 0, 0, 0,-1],
                   [ 0, 0, 1, 0],
                   [ 0, 1, 0, 0],
                   [-1, 0, 0, 0]], dtype=complex) #yapf: disable
    m = rho @ yy @ rho.conj() @ yy
    eigs = [np.abs(e) for e in np.linalg.eig(m)[0]]
    e = [np.sqrt(x) for x in sorted(eigs, reverse=True)]
    return max(0, e[0] - e[1] - e[2] - e[3])


def eof(rho: np.ndarray) -> float:
    """Entanglement of formation of a two-qubit density matrix.
    see http://qwiki.stanford.edu/wiki/Entanglement_of_Formation
    """

    def h(x):
        if x <= 0 or x >= 1:
            return 0
        return -x * np.log2(x) - (1 - x) * np.log2(1 - x)

    C = concurrence(rho)
    arg = max(0, np.sqrt(1 - C**2))
    return h((1 + arg) / 2.0)


def randomUnitary(N: int) -> np.ndarray:
    """Random unitary matrix of dimension N."""
    H = np.random.randn(N, N) + 1j * np.random.randn(N, N)
    H = (H + dagger(H)) / 2
    U = expm(-1j * H)
    return U


def randomState(N: int) -> np.ndarray:
    """Generates a random pure state."""
    psi = np.random.randn(N) + 1j * np.random.randn(N)
    return normalize(psi)


def randomDensity(N: int, rank: Optional[int] = None) -> np.ndarray:
    """Generates a random density matrix.  The distribution does not
    necessarily mean anything.  N is the dimension and rank is the
    number of non-zero eigenvalues."""

    if rank is None or rank > N:
        rank = N
    A = np.random.randn(N, rank) + 1j * np.random.randn(N, rank)
    rho = A @ A.T.conj()
    return rho / np.trace(rho)


##########################################################


def rabi(t, TR, Omega, A, offset):
    return oscillation(t, [(1, Omega)], A, offset) * decay(t, [TR])


def cpmg(t, T1, Tphi, Delta, A, offset, phi=0):
    return oscillation(t, [(np.exp(1j * phi), Delta)], A, offset) * decay(
        t, [2 * T1, Tphi])


def relaxation(t, T1, A, offset):
    return A * decay(t, [T1]) + offset



import bisect
import inspect
import warnings
from abc import ABC, abstractclassmethod
from concurrent.futures import Future
from dataclasses import dataclass, field
from datetime import datetime
from graphlib import TopologicalSorter
from itertools import chain, count
from queue import Empty, Queue
from typing import Any, Callable, Iterable, Sequence, Type

_NODEFAULT = object()


class BaseOptimizer(ABC):

    @abstractclassmethod
    def ask(self) -> tuple:
        pass

    @abstractclassmethod
    def tell(self, suggested: Sequence, value: Any):
        pass

    @abstractclassmethod
    def get_result(self):
        pass


@dataclass
class OptimizerConfig():
    cls: Type[BaseOptimizer]
    dimensions: list = field(default_factory=list)
    args: tuple = ()
    kwds: dict = field(default_factory=dict)
    max_iters: int = 100


class FeedbackPipe():
    __slots__ = (
        'keys',
        '_queue',
    )

    def __init__(self, keys):
        self.keys = keys
        self._queue = Queue()

    def __iter__(self):
        while True:
            try:
                yield self._queue.get_nowait()
            except Empty:
                break

    def __call__(self):
        return self.__iter__()

    def send(self, obj):
        self._queue.put(obj)

    def __repr__(self):
        if not isinstance(self.keys, tuple):
            return f'FeedbackProxy({repr(self.keys)})'
        else:
            return f'FeedbackProxy{self.keys}'


class FeedbackProxy():

    def feedback(self, keywords, obj, suggested=None):
        if keywords in self._pipes:
            if suggested is None:
                suggested = [self.kwds[k] for k in keywords]
            self._pipes[keywords].send((suggested, obj))
        else:
            warnings.warn(f'No feedback pipe for {keywords}', RuntimeWarning,
                          2)

    def feed(self, obj, **options):
        for tracker in self._trackers:
            tracker.feed(self, obj, **options)

    def store(self, obj, **options):
        self.feed(obj, store=True, **options)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_pipes']
        del state['_trackers']
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._pipes = {}
        self._trackers = []


@dataclass
class StepStatus(FeedbackProxy):
    iteration: int = 0
    pos: tuple = ()
    index: tuple = ()
    kwds: dict = field(default_factory=dict)
    vars: list[str] = field(default=list)

    _pipes: dict = field(default_factory=dict, repr=False)
    _trackers: list = field(default_factory=list, repr=False)


@dataclass
class Begin(FeedbackProxy):
    level: int = 0
    iteration: int = 0
    pos: tuple = ()
    index: tuple = ()
    kwds: dict = field(default_factory=dict)
    vars: list[str] = field(default=list)

    _pipes: dict = field(default_factory=dict, repr=False)
    _trackers: list = field(default_factory=list, repr=False)

    def __repr__(self):
        return f'Begin(level={self.level}, kwds={self.kwds}, vars={self.vars})'


@dataclass
class End(FeedbackProxy):
    level: int = 0
    iteration: int = 0
    pos: tuple = ()
    index: tuple = ()
    kwds: dict = field(default_factory=dict)
    vars: list[str] = field(default=list)

    _pipes: dict = field(default_factory=dict, repr=False)
    _trackers: list = field(default_factory=list, repr=False)

    def __repr__(self):
        return f'End(level={self.level}, kwds={self.kwds}, vars={self.vars})'


class Tracker():

    def init(self, loops: dict):
        pass

    def update(self, kwds: dict):
        return kwds

    def feed(self, step: StepStatus, obj: Any, **options):
        pass


def _call_func_with_kwds(func, args, kwds):
    sig = inspect.signature(func)
    for p in sig.parameters.values():
        if p.kind == p.VAR_KEYWORD:
            return func(*args, **kwds)
    kw = {
        k: v
        for k, v in kwds.items()
        if k in list(sig.parameters.keys())[len(args):]
    }
    try:
        return func(*args, **kw)
    except:
        print(args, kw, kwds, sig.parameters.keys())
        raise


def _try_to_call(x, args, kwds):
    if callable(x):
        return _call_func_with_kwds(x, args, kwds)
    return x


def _get_current_iters(loops, level, kwds, pipes):
    keys, current = loops[level]
    limit = -1

    if isinstance(keys, str):
        keys = (keys, )
        current = (current, )
    elif isinstance(keys, tuple) and isinstance(
            current, tuple) and len(keys) == len(current):
        keys = tuple(k if isinstance(k, tuple) else (k, ) for k in keys)
    elif isinstance(keys, tuple) and not isinstance(current, tuple):
        current = (current, )
        if isinstance(keys[0], str):
            keys = (keys, )
    else:
        raise TypeError(f'Illegal keys {keys} on level {level}.')

    if not isinstance(keys, tuple):
        keys = (keys, )
    if not isinstance(current, tuple):
        current = (current, )

    iters = []
    for k, it in zip(keys, current):
        pipe = FeedbackPipe(k)
        if isinstance(it, OptimizerConfig):
            if limit < 0 or limit > it.max_iters:
                limit = it.max_iters
            it = it.cls(it.dimensions, *it.args, **it.kwds)
        else:
            it = iter(_try_to_call(it, (), kwds))

        iters.append((it, pipe))
        pipes[k] = pipe

    return keys, iters, pipes, limit


def _generate_kwds(keys, iters, kwds, iteration, limit):
    ret = {}
    for ks, it in zip(keys, iters):
        if isinstance(ks, str):
            ks = (ks, )
        if hasattr(it[0], 'ask') and hasattr(it[0], 'tell') and hasattr(
                it[0], 'get_result'):
            if limit > 0 and iteration >= limit - 1:
                value = _call_func_with_kwds(it[0].get_result, (), kwds).x
            else:
                value = _call_func_with_kwds(it[0].ask, (), kwds)
        else:
            value = next(it[0])
            if len(ks) == 1:
                value = (value, )
        ret.update(zip(ks, value))
    return ret


def _send_feedback(generator, feedback):
    if hasattr(generator, 'ask') and hasattr(generator, 'tell') and hasattr(
            generator, 'get_result'):
        generator.tell(*feedback)


def _feedback(iters):
    for generator, pipe in iters:
        for feedback in pipe():
            _send_feedback(generator, feedback)


def _call_functions(functions, kwds, order):
    vars = []
    for i, k in enumerate(order):
        if k in kwds:
            continue
        elif k in functions:
            kwds[k] = _try_to_call(functions[k], (), kwds)
            vars.append(k)
        else:
            break
    else:
        return [], vars
    return order[i:], vars


def _args_generator(loops: list, kwds: dict[str,
                                            Any], level: int, pos: tuple[int,
                                                                         ...],
                    vars: list[tuple[str]], filter: Callable[..., bool] | None,
                    functions: dict[str, Callable], trackers: list[Tracker],
                    pipes: dict[str | tuple[str, ...],
                                FeedbackPipe], order: list[str]):
    order, local_vars = _call_functions(functions, kwds, order)
    if len(loops) == level and level > 0:
        if order:
            raise TypeError(f'Unresolved functions: {order}')
        for tracker in trackers:
            kwds = tracker.update(kwds)
        if filter is None or _call_func_with_kwds(filter, (), kwds):
            yield StepStatus(
                pos=pos,
                kwds=kwds,
                vars=[*vars[:-1], tuple([*vars[-1], *local_vars])],
                _pipes=pipes,
                _trackers=trackers)
        return

    keys, current_iters, pipes, limit = _get_current_iters(
        loops, level, kwds, pipes)

    for i in count():
        if limit > 0 and i >= limit:
            break
        try:
            kw = _generate_kwds(keys, current_iters, kwds, i, limit)
        except StopIteration:
            break
        yield Begin(level=level,
                    pos=pos + (i, ),
                    kwds=kwds | kw,
                    vars=[*vars, tuple([*local_vars, *kw.keys()])],
                    _pipes=pipes,
                    _trackers=trackers)
        yield from _args_generator(
            loops, kwds | kw, level + 1, pos + (i, ),
            [*vars, tuple([*local_vars, *kw.keys()])], filter, functions,
            trackers, pipes, order)
        yield End(level=level,
                  pos=pos + (i, ),
                  kwds=kwds | kw,
                  vars=[*vars, tuple([*local_vars, *kw.keys()])],
                  _pipes=pipes,
                  _trackers=trackers)
        _feedback(current_iters)


def _find_diff_pos(a: tuple, b: tuple):
    for i, (x, y) in enumerate(zip(a, b)):
        if x != y:
            return i


def _build_dependence(loops, functions, constants):
    graph = {}
    loop_names = set()
    var_names = set()
    for keys in loops:
        level_vars = set()
        if isinstance(keys, str):
            keys = (keys, )
        for ks in keys:
            if isinstance(ks, str):
                ks = (ks, )
            level_vars.update(ks)
            for k in ks:
                graph.setdefault(k, set()).update(loop_names)
        loop_names.update(level_vars)
        var_names.update(level_vars)
    var_names.update(functions.keys())
    var_names.update(constants.keys())

    def _add_dependence(keys, value):
        if isinstance(keys, str):
            keys = (keys, )
        for key in keys:
            graph.setdefault(key, set())
            for k, p in inspect.signature(value).parameters.items():
                if p.kind in [
                        p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD,
                        p.KEYWORD_ONLY
                ] and k in var_names:
                    graph[key].add(k)
                if p.kind == p.VAR_KEYWORD and key not in loop_names:
                    graph[key].update(loop_names)

    for keys, values in chain(loops.items(), functions.items()):
        if callable(values):
            _add_dependence(keys, values)
        elif isinstance(values, tuple):
            for ks, v in zip(keys, values):
                if callable(v):
                    _add_dependence(ks, v)

    return graph


def scan_iters(loops: dict[str | tuple[str, ...],
                           Iterable | Callable | OptimizerConfig
                           | tuple[Iterable | Callable | OptimizerConfig,
                                   ...]] = {},
               filter: Callable[..., bool] | None = None,
               functions: dict[str, Callable] = {},
               constants: dict[str, Any] = {},
               trackers: list[Tracker] = [],
               level_marker: bool = False,
               **kwds) -> Iterable[StepStatus]:
    """
    Scan the given iterable of iterables.

    Parameters
    ----------
    loops : dict
        A map of iterables that are scanned.
    filter : Callable[..., bool]
        A filter function that is called for each step.
        If it returns False, the step is skipped.
    functions : dict
        A map of functions that are called for each step.
    constants : dict
        Additional keyword arguments that are passed to the iterables.

    Returns
    -------
    Iterable[StepStatus]
        An iterable of StepStatus objects.

    Examples
    --------
    >>> iters = {
    ...     'a': range(2),
    ...     'b': range(3),
    ... }
    >>> list(scan_iters(iters))
    [StepStatus(iteration=0, pos=(0, 0), index=(0, 0), kwds={'a': 0, 'b': 0}),
     StepStatus(iteration=1, pos=(0, 1), index=(0, 1), kwds={'a': 0, 'b': 1}),
     StepStatus(iteration=2, pos=(0, 2), index=(0, 2), kwds={'a': 0, 'b': 2}),
     StepStatus(iteration=3, pos=(1, 0), index=(1, 0), kwds={'a': 1, 'b': 0}),
     StepStatus(iteration=4, pos=(1, 1), index=(1, 1), kwds={'a': 1, 'b': 1}),
     StepStatus(iteration=5, pos=(1, 2), index=(1, 2), kwds={'a': 1, 'b': 2})]

    >>> iters = {
    ...     'a': range(2),
    ...     'b': range(3),
    ... }
    ... list(scan_iters(iters, lambda a, b: a < b))
    [StepStatus(iteration=0, pos=(0, 1), index=(0, 0), kwds={'a': 0, 'b': 1}),
     StepStatus(iteration=1, pos=(0, 2), index=(0, 1), kwds={'a': 0, 'b': 2}),
     StepStatus(iteration=2, pos=(1, 2), index=(1, 0), kwds={'a': 1, 'b': 2})]
    """

    # TODO: loops 里的 callable 值如果有 VAR_KEYWORD 参数，并且在运行时实际依
    #       赖于 functions 里的某些值，则会导致依赖关系错误
    # TODO: functions 里的 callable 值如果有 VAR_KEYWORD 参数，则对这些参数
    #       的依赖会被认为是对全体循环参数的依赖，并且这些函数本身不存在相互依赖

    if 'additional_kwds' in kwds:
        functions = functions | kwds['additional_kwds']
    if 'iters' in kwds:
        loops = loops | kwds['iters']

    if len(loops) == 0:
        return

    for tracker in trackers:
        tracker.init(loops)

    graph = _build_dependence(loops, functions, constants)
    ts = TopologicalSorter(graph)
    order = []
    ts.prepare()
    while ts.is_active():
        ready = ts.get_ready()
        for k in ready:
            ts.done(k)
        order.append(ready)
    print(order)

    order = list(TopologicalSorter(graph).static_order())

    last_step = None
    index = ()
    iteration = count()

    for step in _args_generator(list(loops.items()),
                                kwds=constants,
                                level=0,
                                pos=(),
                                vars=[],
                                filter=filter,
                                functions=functions,
                                trackers=trackers,
                                pipes={},
                                order=order):
        if isinstance(step, (Begin, End)):
            if level_marker:
                if last_step is not None:
                    step.iteration = last_step.iteration
                yield step
            continue

        if last_step is None:
            i = 0
            index = (0, ) * len(step.pos)
        else:
            i = _find_diff_pos(last_step.pos, step.pos)
            index = tuple((j <= i) * n + (j == i) for j, n in enumerate(index))
        step.iteration = next(iteration)
        step.index = index
        yield step
        last_step = step


class Storage(Tracker):
    """
    A tracker that stores the results of the steps.

    Parameters
    ----------
    storage : dict
        The storage of the results.
    shape : tuple
        The shape of the results.
    ctime : datetime.datetime
        The creation time of the tracker.
    mtime : datetime.datetime
        The modification time of the tracker.
    """

    def __init__(self,
                 storage: dict = None,
                 shape: tuple = (),
                 save_kwds: bool | Sequence[str] = True,
                 frozen_keys: tuple = ()):
        self.ctime = datetime.utcnow()
        self.mtime = datetime.utcnow()
        self.storage = storage if storage is not None else {}
        self.cache = {}
        self.pos = {}
        self.timestamps = {}
        self.iteration = {}
        self._init_keys = list(self.storage.keys())
        self._frozen_keys = frozen_keys
        self._key_levels = ()
        self.shape = shape
        self.count = 0
        self.save_kwds = save_kwds
        self.queue = Queue()
        self._queue_buffer = None

    def init(self, loops: dict):
        """
        Initialize the tracker.

        Parameters
        ----------
        loops : dict
            The map of iterables.
        """
        from numpy import ndarray

        for level, (keys, iters) in enumerate(loops.items()):
            self._key_levels = self._key_levels + ((keys, level), )
            if isinstance(keys, str):
                keys = (keys, )
                iters = (iters, )
            if (len(keys) > 1 and len(iters) == 1
                    and isinstance(iters[0], ndarray) and iters[0].ndim == 2
                    and iters[0].shape[1] == len(keys)):
                iters = iters[0]
                for i, key in enumerate(keys):
                    self.storage[key] = iters[:, i]
                    self._frozen_keys = self._frozen_keys + (key, )
                    self._init_keys.append(key)
                continue
            if not isinstance(iters, tuple) or len(keys) != len(iters):
                continue
            for key, iter in zip(keys, iters):
                if key not in self.storage and isinstance(
                        iter, (list, range, ndarray)):
                    self.storage[key] = iter
                    self._frozen_keys = self._frozen_keys + (key, )
                    self._init_keys.append(key)

    def feed(self, step: StepStatus, dataframe: dict, store=False, **options):
        """
        Feed the results of the step to the storage.

        Parameters
        ----------
        step : StepStatus
            The step.
        dataframe : dict
            The results of the step.
        """
        import numpy as np

        if not store:
            return
        self.mtime = datetime.utcnow()
        if not self.shape:
            self.shape = tuple([i + 1 for i in step.pos])
        else:
            self.shape = tuple(
                [max(i + 1, j) for i, j in zip(step.pos, self.shape)])
        if self.save_kwds:
            if isinstance(self.save_kwds, bool):
                kwds = step.kwds
            else:
                kwds = {
                    key: step.kwds.get(key, np.nan)
                    for key in self.save_kwds
                }
        else:
            kwds = {}
        self.queue.put_nowait(
            (step.iteration, step.pos, dataframe, kwds, self.mtime))

    def _append(self, iteration, pos, dataframe, kwds, now):
        for k, v in chain(kwds.items(), dataframe.items()):
            if k in self._frozen_keys:
                continue
            self.count += 1
            if k not in self.storage:
                self.storage[k] = [v]
                self.pos[k] = tuple([i] for i in pos)
                self.timestamps[k] = [now.timestamp()]
                self.iteration[k] = [iteration]
            else:
                self.storage[k].append(v)
                for i, l in zip(pos, self.pos[k]):
                    l.append(i)
                self.timestamps[k].append(now.timestamp())
                self.iteration[k].append(iteration)

    def _flush(self, block=False):
        if self._queue_buffer is not None:
            iteration, pos, fut, kwds, now = self._queue_buffer
            if fut.done() or block:
                self._append(iteration, pos, fut.result(), kwds, now)
                self._queue_buffer = None
            else:
                return
        while not self.queue.empty():
            iteration, pos, dataframe, kwds, now = self.queue.get()
            if isinstance(dataframe, Future):
                if not dataframe.done() and not block:
                    self._queue_buffer = (iteration, pos, dataframe, kwds, now)
                    return
                else:
                    self._append(iteration, pos, dataframe.result(), kwds, now)
            else:
                self._append(iteration, pos, dataframe, kwds, now)

    def _get_array(self, key, shape, count):
        import numpy as np

        data, data_shape, data_count = self.cache.get(key, (None, (), 0))
        if (data_shape, data_count) == (shape, count):
            return data
        tmp = np.asarray(self.storage[key])
        if data_shape != shape:
            data = np.full(shape + tmp.shape[1:], np.nan, dtype=tmp.dtype)
        data[self.pos[key]] = tmp
        self.cache[key] = (data, shape, count)
        return data

    def _get_part(self, key, skip):
        i = bisect.bisect_left(self.iteration[key], skip)
        pos = tuple(p[i:] for p in self.pos[key])
        iteration = self.iteration[key][i:]
        data = self.storage[key][i:]
        return data, iteration, pos

    def keys(self):
        """
        Get the keys of the storage.
        """
        self._flush()
        return self.storage.keys()

    def values(self):
        """
        Get the values of the storage.
        """
        self._flush()
        return [self[k] for k in self.storage]

    def items(self):
        """
        Get the items of the storage.
        """
        self._flush()
        return list(zip(self.keys(), self.values()))

    def get(self, key, default=_NODEFAULT, skip=None, block=False):
        """
        Get the value of the storage.
        """
        self._flush(block)
        if key in self._init_keys:
            return self.storage[key]
        elif key in self.storage:
            if skip is None:
                return self._get_array(key, self.shape, self.count)
            else:
                return self._get_part(key, skip)
        elif default is _NODEFAULT:
            raise KeyError(key)
        else:
            return default

    def __getitem__(self, key):
        return self.get(key)

    def __getstate__(self):
        self._flush()
        storage = dict(self.items())
        return {
            'storage': storage,
            'pos': self.pos,
            'timestamps': self.timestamps,
            'iteration': self.iteration,
            'shape': self.shape,
            'ctime': self.ctime,
            'mtime': self.mtime,
            '_init_keys': self._init_keys,
            '_frozen_keys': self._frozen_keys,
            '_key_levels': self._key_levels,
            'save_kwds': self.save_kwds,
        }
